"""This is a module that handles the graphical user interface provided by Ogre through CEGUI.
http://www.cegui.org.uk/wiki/index.php/Falagard_System_base_widgets_reference
"""
__version__ = "$Revision: 1.13 $"
__source__ = "$Source: /cvsroot/game-client-system/scripts/GUI.py,v $"
__author__ = "Kjartan A Jonsson"
__copyright__ = "Copyright (c) 2007 aGameCompany"
__license__ = "aGameCompany"
import gameGUI


class GUIError(Exception): pass

class SystemGUI :
    
    __GUIProperty_Text = 4
    __GUIProperty_Property = 5
    __GUIProperty_ItemAdd = 6
    __GUIProperty_ItemsClear = 7
    __GUIProperty_ItemsSelected = 8
    __GUIProperty_ImageSet_Load = 9
    __GUIProperty_CheckboxSelected = 10
    __GUIProperty_Call = 11
    __GUIProperty_Position = 12

    def loadLayout (self, fileName, w=256, h=256) :
        """
        Loads a legal CEGUI formated layout where fileName is name of the file and  
        must exist be accessable in resource paths.
        Passing no fileName will unload the previusly loaded layout.
        """
        res = gameGUI.loadGUI (fileName, w, h)
        if res <> None : raise GUIError, "Could not load layout got: "+ str(res)

    def addLayout (self, fileName) :
        """
        Loads a legal CEGUI formated layout where fileName is name of the file and  
        must exist be accessable in resource paths.
        Passing no fileName will unload the previusly loaded layout.
        """
        res = gameGUI.addGUI (fileName)
        if res <> None : raise GUIError, "Could not load layout got: "+ str(res)
        
    def delLayout (self, windowName) :
        """
        Loads a legal CEGUI formated layout where fileName is name of the file and  
        must exist be accessable in resource paths.
        Passing no fileName will unload the previusly loaded layout.
        """
        res = gameGUI.delGUI (windowName)        
        if res <> None : raise GUIError, "Could not delete layout got: "+ str(res)

    def loadImageSet (self, fileName) :
        """
        Loads an image set from file. The fileName is the name of a file loaded through the ogre resource manager.
        If the imageset is loaded we dont do anything. If it is not found we notify failure.
        Raises GUIError on failure.
        """
        if fileName == "" : raise GUIError, "Could not load image set since it is empty ('')"
        self.__setProperty(self.__GUIProperty_ImageSet_Load, {"ImageSetFile":fileName})

    def addEventSubscriber (self, windowName, eventName, funcName, funcArgs) :
        """
        Assigns a event subscriber for the widget for a layout must have been successfully been loaded through call loadGUI()
        where windowName is a window that exists in the loaded layout. eventName is a legal event name (see note bellow).
        functionName is a python string that will be executed on call. This is must be a defined function or 
        method. This must be accessable from the __init__.py file meaning that if the call is callable from the setup
        script then it will be successfully called (raises an python error otherwise)
        args is the python (functionName) method or function arguments passed to the function. Can be dictionary string.
        Raises GUIError on failure.
        The type of events are dependent on the widget called. Subscribing an event to a widget that does not support it will
        crash the system. To get a list of events supported you can look at the following links. The public attributes has a list of events
        that will be called. Usually you can use the name preciding 'Event'. For instance button has an name EventClicked, you call with event
        name being 'Clicked'. 
        button (http://www.cegui.org.uk/api_reference/classCEGUI_1_1PushButton.html#65d6b712c449f7a9244dcc9d40c0d825)
        window (http://www.cegui.org.uk/api_reference/classCEGUI_1_1Window.html)
        editbox (http://www.cegui.org.uk/api_reference/classCEGUI_1_1Editbox.html)
        Bug: Subscribing an event to a widget that does not support it will crash the system
        """
        if (windowName == "") : raise GUIError, "Could not add subscriber to window. Because windowName was empty ''"
        if (eventName == "") : raise GUIError, "Could not add subscriber to window. Because eventName was empty ''"
        if (funcName == "") : raise GUIError, "Could not add subscriber to window. Because funcName was empty ''"
        if gameGUI.createGUIEventSubscriber (windowName, eventName, funcName, str(funcArgs).replace("'","\"")) == 0 : 
            raise GUIError, "Could not add subscriber to window '"+windowName+"' for this layout. Probably because the window does not exist."
    
    def setText (self, windowName, text) :
        """
        Sets a gui text value in a loaded layout having window with name windowName.
        Raises GUIError if windowName does not exist in current layout.
        """
        self.__setProperty(self.__GUIProperty_Text, {"Name":windowName,"Text":text})

    def setPosition (self, windowName, x, y) :
        """
        """
        self.__setProperty(self.__GUIProperty_Position, {"Name":windowName,"x":x, "y":y})

    def getText (self, windowName) :
        """
        Gets a gui text value in a loaded layout having window with name windowName.
        Raises GUIError if windowName does not exist in current layout.
        """
        return self.__getProperty(self.__GUIProperty_Text, {"Name":windowName})
    
    def setProperty (self, windowName, PropertyName, PropertyValue):
        """
        Direct access to gui widgets window property in a loaded layout. 
        Refer to CEGUI reference at http://www.cegui.org.uk/api_reference/
        or http://www.cegui.org.uk/api_reference/namespaceCEGUI_1_1WindowProperties.html
        Raises GUIError if windowName does not exist in current layout or property is 
        not present in this widget.
        
        For Hikari
        	play (start, stop, loop, rewind)
        	transparent (bool) : uses alpha channel
        	show (bool) : hides/shows window
        	opacity (real) : sets window opacity
        	
        """
        self.__setProperty(self.__GUIProperty_Property, {"Name":windowName,"Property":PropertyName, "Value":PropertyValue})

    def getProperty (self, windowName, PropertyName):
        """
        Direct access to gui widgets window property in a loaded layout. 
        Refer to CEGUI reference at http://www.cegui.org.uk/api_reference/
        or http://www.cegui.org.uk/api_reference/namespaceCEGUI_1_1WindowProperties.html
        Raises GUIError if windowName does not exist in current layout or property is 
        not present in this widget.
        """
        return self.__getProperty(self.__GUIProperty_Property, {"Name":windowName,"Property":PropertyName})
    
    def call (self, windowName, func, arg) :
        """
        Used in Hikari interface to forward the call to the flash window. Returns None on failure.
        Where func is the defined external function in the flash control and arg is a string.
        """
        return gameGUI.GUI_setProperty (self.__GUIProperty_Call, str({"Name":windowName,"Property":func, "Value":arg}))
    
    def addItem (self, windowName, itemName) :
        """
        Appends item last to a window. Widget exists and is of type TaharezLook/Combobox, TaharezLook/Listbox
        Item is added to the widget, windowName is the name of the widget and itemName is the name 
        of the item added (as displayed in window). Raises GUIError on failure       
        """
        self.__setProperty(self.__GUIProperty_ItemAdd, {"Name":windowName,"ItemName":itemName})

    def clearItems (self, windowName) :
        """
        Clears items from a window. Widget exists and is of type TaharezLook/Combobox, TaharezLook/Listbox
        Item is added to the widget, windowName is the name of the widget. Raises GUIError on failure
        """
        self.__setProperty(self.__GUIProperty_ItemsClear, {"Name":windowName})

    def getSelectedItems (self, windowName):
        """
        Returns selected items from a window. Widget exists and is of type TaharezLook/Combobox, TaharezLook/Listbox
        The selected items are returned in a list. Raises GUIError on failure
        """
        filtered = str(self.__getProperty(self.__GUIProperty_ItemsSelected, {"Name":windowName}))
        filtered = filtered.replace("[,","[")
        return eval(filtered)
        
    def getCheckboxSelected (self, windowName):
        """
        Returns selected items from a window. Widget exists and is of type TaharezLook/Checkbox
        Returns True/False. Raises GUIError on failure
        """
        filtered = str(self.__getProperty(self.__GUIProperty_CheckboxSelected, {"Name":windowName}))
        return eval(filtered)
   
    def __setProperty (self, property, attrib) :
        """
        Internal function setting the gui widget property. Raises GUIError if property trap
        does not exist in the system or the attributes in attrib was not passed correctly.
        """
        error = gameGUI.GUI_setProperty (property, str(attrib))
        if error <> None: 
            raise GUIError, "Error calling __setProperty " + str(property) + " and attrib " + str(attrib) + ". Got error :" + error

    def __getProperty (self, property, attrib) :
        """
        Internal function setting the gui widget property. Raises GUIError if property trap
        does not exist in the system or the attributes in attrib was not passed correctly.
        """
        res = gameGUI.GUI_getProperty (property, str(attrib))
        if res == None:  raise GUIError, "Error calling __getProperty " + str(property) + "and attrib " + str(attrib)
        return res

systemGUI = SystemGUI()
