"""This module behaves as a singleton allowing calls to the game physics system"""
__author__ = "Kjartan A Jonsson"
__version__ = "$Revision: 1.41 $"
__copyright__ = "Copyright (c) 2007 aGameCompany"
__license__ = "aGameCompany"

import gamePhysics
import gameGraphics
import SceneManager
from validatevalue import *

#general
__PhysicsWorldProperty_Destroy = 0
__PhysicsWorldProperty_Setup = 1
__PhysicsWorldProperty_Debug = 2
__PhysicsWorldProperty_Area = 3
__PhysicsWorldProperty_Marker = 4
#body
__PhysicsBodyProperty_Destroy = 0
__PhysicsBodyProperty_Box = 1
__PhysicsBodyProperty_Ball = 2
__PhysicsBodyProperty_Capsule = 3
__PhysicsBodyProperty_Convex = 4 
__PhysicsBodyProperty_Prism = 5
__PhysicsBodyProperty_Position = 6
__PhysicsBodyProperty_Orientation = 7 
#vehicle
__PhysicsVehicleProperty_Destroy = 0
__PhysicsVehicleProperty_Create = 1
__PhysicsVehicleProperty_Input = 2
__PhysicsVehicleProperty_Position = 3
__PhysicsVehicleProperty_Orientation = 4
__PhysicsVehicleProperty_AddWheel = 5
#tools
__PhysicsToolProperty_Destroy = 0
__PhysicsToolProperty_Gripper = 1
__PhysicsToolProperty_Magnet = 2
__PhysicsToolProperty_Gun = 3
__PhysicsToolProperty_Rocket = 4


__bodies = {}
__vehicles = {}
__tools = {}

class PhysicsError(Exception): pass
class BodyError (PhysicsError) : pass
class VehicleError (PhysicsError) : pass
class ToolError (PhysicsError) : pass

class Tool :
    __PhysicsToolProperty_Enable = 21 #starts of the rockets, fires the gun, opens/closes the gripper and turns magnet on/off
    __PhysicsToolProperty_Action = 22 #starts of the rockets, fires the gun, opens/closes the gripper and turns magnet on/off

    __PhysicsToolProperty_PreEffectSuccess = 51
    __PhysicsToolProperty_PreEffectFailure = 52
    __PhysicsToolProperty_NeoEffect = 53
    __PhysicsToolProperty_BallisticsEffect = 54
    __PhysicsToolProperty_PostEffect = 55
    __PhysicsToolProperty_RechargeEffect = 56
    __PhysicsToolProperty_DefaultEffect = 57
		#from body
    __PhysicsBodyProperty_Data = 22
    __PhysicsBodyProperty_Attach = 31


    name = None
    id = None
    file = None #mesh file used by this body
    def __init__ (self, dic) :
        """class initalization call that raises an error when called directly"""
        self.id = self.id = dic["Name"]
        self.file = dic["MeshFile"]

    def setEnabled (self, b) :
        _setToolProperty (self.__PhysicsToolProperty_Enable, {"Name":self.id, "Enable":str(b).lower()})

    def setAction (self, a) :
        """
        Where a is an integer action type that is know by the tool
        Common values are in the range of -1 to 1
        """    	
        _setToolProperty (self.__PhysicsToolProperty_Action, {"Name":self.id, "Action":a})

    def attach (self, body, pa=[], pb=[]) :
        """Will attach body to this body pa, pb is the delta from center of object"""
        pass
        #_setToolProperty (self.__PhysicsBodyProperty_Attach, {"Name":self.id,"RotationW":w,"RotationX":x,"RotationY":y,"RotationZ":z})

    def setPreEffectSuccess (self, anim=None, sound=None, psys=None, ppos=None, msec=0) :
        """
        Sets the pre effect when the tool successfully functions
        Where :
        	anim is name of the animation that this entity body currently has defined, returns error if not found. Default is None not used.
        	sound is WeaponSound of the sound file to be played on success, returns error if not found. Default is None not used.
        	psys is the name of the defined particle system, returns error if not found. Default is None not used.
        	ppos is the position of the particle system in local coordinates (body position is null). Default is None not used.
        	sec is tool effect time and is 0 by default.
        This functions allows total control of what effect to be played out when the tool is used.
        """
        ppos = ""
        dic = {"Name":self.id, "WeaponAnimation":anim, "WeaponSound":sound, "WeaponParticlesystem":psys, "WeaponParticlesystemPosition":ppos, "WeaponEffectTime":msec}
        _setToolProperty (self.__PhysicsToolProperty_PreEffectSuccess, dic)

    def setPreEffectFailure (self, anim=None, sound=None, psys=None, ppos=None, msec=0) :
        """
        Sets the pre effect when the tool does not function due to some internal state (missing bullets or power less).
        Where :
        	anim is name of the animation that this entity body currently has defined, returns error if not found. Default is None not used.
        	sound is WeaponSound of the sound file to be played on success, returns error if not found. Default is None not used.
        	psys is the name of the defined particle system, returns error if not found. Default is None not used.
        	ppos is the position of the particle system in local coordinates (body position is null). Default is None not used.
        	sec is tool effect time and is 0 by default.
        This functions allows total control of what effect to be played out when the tool is used.
        """
        ppos = ""
        dic = {"Name":self.id, "WeaponAnimation":anim, "WeaponSound":sound, "WeaponParticlesystem":psys, "WeaponParticlesystemPosition":ppos, "WeaponEffectTime":msec}
        _setToolProperty (self.__PhysicsToolProperty_PreEffectFailure, dic)

    def setNeoEffect (self, anim=None, sound=None, psys=None, ppos=None, msec=0) :
        """
        Sets the neo (new ballistics or creation of) effect when the tool does executes the tool (more specifically after pre msec)
        Where :
        	anim is name of the animation that this entity body currently has defined, returns error if not found. Default is None not used.
        	sound is WeaponSound of the sound file to be played on success, returns error if not found. Default is None not used.
        	psys is the name of the defined particle system, returns error if not found. Default is None not used.
        	ppos is the position of the particle system in local coordinates (body position is null). Default is None not used.
        	sec is tool effect time and is 0 by default.
        This functions allows total control of what effect to be played out when the tool is used.
        """
        ppos = ""
        dic = {"Name":self.id, "WeaponAnimation":anim, "WeaponSound":sound, "WeaponParticlesystem":psys, "WeaponParticlesystemPosition":ppos, "WeaponEffectTime":msec}
        _setToolProperty (self.__PhysicsToolProperty_NeoEffect, dic)
        
    def setPostEffect (self, anim=None, sound=None, psys=None, ppos=None, msec=0) :
        """
        Sets the post effect when the tool has finished executing the neo effect (more specifically after neo msec).
        Where :
        	anim is name of the animation that this entity body currently has defined, returns error if not found. Default is None not used.
        	sound is WeaponSound of the sound file to be played on success, returns error if not found. Default is None not used.
        	psys is the name of the defined particle system, returns error if not found. Default is None not used.
        	ppos is the position of the particle system in local coordinates (body position is null). Default is None not used.
        	sec is tool effect time and is 0 by default.
        This functions allows total control of what effect to be played out when the tool is used.
        """
        ppos = ""
        dic = {"Name":self.id, "WeaponAnimation":anim, "WeaponSound":sound, "WeaponParticlesystem":psys, "WeaponParticlesystemPosition":ppos, "WeaponEffectTime":msec}
        _setToolProperty (self.__PhysicsToolProperty_PostEffect, dic)
        
    def setRechargeEffect (self, anim=None, sound=None, psys=None, ppos=None, msec=0) :
        """
        Sets the pre effect when the tool does not function due to some internal state (missing bullets or power less).
        Where :
        	anim is name of the animation that this entity body currently has defined, returns error if not found. Default is None not used.
        	sound is WeaponSound of the sound file to be played on success, returns error if not found. Default is None not used.
        	psys is the name of the defined particle system, returns error if not found. Default is None not used.
        	ppos is the position of the particle system in local coordinates (body position is null). Default is None not used.
        	sec is tool effect time and is 0 by default.
        This functions allows total control of what effect to be played out when the tool is used.
        Neither effects can be running when this is played
        """
        ppos = ""
        dic = {"Name":self.id, "WeaponAnimation":anim, "WeaponSound":sound, "WeaponParticlesystem":psys, "WeaponParticlesystemPosition":ppos, "WeaponEffectTime":msec}
        _setToolProperty (self.__PhysicsToolProperty_RechargeEffect, dic)

    def setDefaultEffect (self, anim=None, sound=None, psys=None, ppos=None, msec=0) :
        """
        Sets the default effect when the tool is idle.
        Where :
        	anim is name of the animation that this entity body currently has defined, returns error if not found. Default is None not used.
        	sound is WeaponSound of the sound file to be played on success, returns error if not found. Default is None not used.
        	psys is the name of the defined particle system, returns error if not found. Default is None not used.
        	ppos is the position of the particle system in local coordinates (body position is null). Default is None not used.
        	sec is tool effect time and is 0 by default.
        This functions allows total control of what effect to be played out when the tool is used.
        """
        ppos = ""
        dic = {"Name":self.id, "WeaponAnimation":anim, "WeaponSound":sound, "WeaponParticlesystem":psys, "WeaponParticlesystemPosition":ppos, "WeaponEffectTime":msec}
        _setToolProperty (self.__PhysicsToolProperty_DefaultEffect, dic)
        
    def setBallisticsEffect (self, entity=None, sound=None, psys=None, ppos=None) :
        """
        Sets the pre effect when the tool does not function due to some internal state (missing bullets or power less).
        Where :
        	entity is the ballistics entity created
        	anim is name of the animation that this entity body currently has defined, returns error if not found. Default is None not used.
        	sound is WeaponSound of the sound file to be played on success, returns error if not found. Default is None not used.
        	psys is the name of the defined particle system, returns error if not found. Default is None not used.
        	ppos is the position of the particle system in local coordinates (body position is null). Default is None not used.
        This functions allows total control of what effect to be played out when the tool is used.
        """
        ppos = ""
        dic = {"Name":self.id, "BallisticEntity":entity, "BallisticAnimation":anim, "BallisticSound":sound, "BallisticParticlesystem":psys, "BallisticParticlesystemPosition":ppos}
        _setToolProperty (self.__PhysicsToolProperty_RechargeEffect, dic)
        

    def getData (self) :
        v = _getBodyProperty (self.__PhysicsBodyProperty_Data,{"Name":self.id})
        if v == "" : raise BodyError, "Could not get body with name "+str(self.id)
        p = eval (v)
        return p

    def getEntity (self) :
        data = self.getData()
        entity = SceneManager.Entity (data["entity"])
        return entity
        
        
class Body (SceneManager.Entity) :
    __PhysicsBodyProperty_Position = 6
    __PhysicsBodyProperty_Orientation = 7
    __PhysicsBodyProperty_Material = 21
    __PhysicsBodyProperty_Data = 22
    __PhysicsBodyProperty_Attach = 31
    
    name = None
    id = None
    file = None #mesh file used by this body
    def __init__ (self, dic) :
        """class initalization call that raises an error when called directly"""
        self.id = self.id = dic["Name"]
        self.file = dic["MeshFile"]

    def getData (self) :
        v = _getBodyProperty (self.__PhysicsBodyProperty_Data,{"Name":self.id})
        if v == "" : raise BodyError, "Could not get body with name "+str(self.id)
        p = eval (v)
        return p

    def getEntity (self) :
        data = self.getData()
        entity = SceneManager.Entity (data["entity"])
        return entity

    def getPosition (self) :
        v = _getBodyProperty (self.__PhysicsBodyProperty_Position,{"Name":self.id})
        p = eval (v.replace(" ", ", "))
        return p

    def getOrientation (self) :
        v = _getBodyProperty (self.__PhysicsBodyProperty_Orientation,{"Name":self.id})
        p = eval (v.replace(" ", ", "))
        return p

    def setPosition (self, x, y, z) :
        _setBodyProperty (self.__PhysicsBodyProperty_Position, {"Name":self.id,"PositionX":x,"PositionY":y,"PositionZ":z})

    def setOrientation (self, w, x, y, z) :
        _setBodyProperty (self.__PhysicsBodyProperty_Orientation, {"Name":self.id,"RotationW":w,"RotationX":x,"RotationY":y,"RotationZ":z})

    def attach(self, body, pos=[0,0,0], lforce=0, aforce=0) : 
        """ 
        Attaches the body to vehicles chassis position in local space where 
        - force is the strenght of the attachement. Force must be positive or 0 value
        - pos is the position and 
        - body is the body instance
        """
        if lforce < 0 : raise VehicleError, "lorce was smaller than 0"
        if aforce < 0 : raise VehicleError, "aforce was smaller than 0"
        if len(pos) != 3 : raise VehicleError, "The position must be a array having 3 real values"
        _setBodyProperty (self.__PhysicsBodyProperty_Attach, {"Name":self.id, "TargetName":body.id, "AngularForce": aforce, "LinearForce": lforce, "PositionX":pos[0], "PositionY":pos[1], "PositionZ":pos[2]})


class Vehicle :
    """
    Is the static and dynamic representation of our vehicle. The class requires calls to setup the vehicle data (will use default data if applicable)
    and then during calls makeDynamic and makeStatic we create the actual vehicle based on this data.
    Vehicle parts are represented using bolts and an vehicle part. If the vehicle mesh is omitted then no mesh and bolds are not created.
    Vehicle parts are connected using bolts if mesh is NULL then no visual representation (entity) is created     
    """
    __PhysicsVehicleProperty_Destroy = 0
    __PhysicsVehicleProperty_Create = 1
    __PhysicsVehicleProperty_Input = 2
    __PhysicsVehicleProperty_Position = 3
    __PhysicsVehicleProperty_Orientation = 4
    __PhysicsVehicleProperty_Chassis = 5
    __PhysicsVehicleProperty_AddWheel = 6
    __PhysicsVehicleProperty_DelWheel = 7
    __PhysicsVehicleProperty_Suspension = 8
    __PhysicsVehicleProperty_EngineSetup = 9
    __PhysicsVehicleProperty_EnginePowerStep = 10
    __PhysicsVehicleProperty_EnginePowerStepClear = 11
    __PhysicsVehicleProperty_WheelTracks = 12
    __PhysicsVehicleProperty_Correction = 13
    __PhysicsVehicleProperty_PositionOrientationExpected = 14
    __PhysicsVehicleProperty_DeadRecon = 15
    __PhysicsVehicleProperty_WheelParticles = 16
    __PhysicsVehicleProperty_Steer = 17
    __PhysicsVehicleProperty_MakeDynamic = 18
    __PhysicsVehicleProperty_AddAxis = 19
    __PhysicsVehicleProperty_AddBrake = 20
    __PhysicsVehicleProperty_MakeStatic = 21
    __PhysicsVehicleProperty_Name = 22
    __PhysicsVehicleProperty_TireGrip = 23
    __PhysicsVehicleProperty_TireBehavior = 24
    __PhysicsVehicleProperty_TireMass = 25
    __PhysicsVehicleProperty_Brake = 26
    __PhysicsVehicleProperty_ControlArms = 27
    __PhysicsVehicleProperty_DriveShaft = 28
    __PhysicsVehicleProperty_Wheel = 29
    __PhysicsVehicleProperty_ShockAbsorbers = 30
    __PhysicsVehicleProperty_BoltFrame = 31
    __PhysicsVehicleProperty_BoltAxis = 32
    __PhysicsVehicleProperty_Attach = 33
    
    __type = {"Free": 0, "Turn":1, "Turn&Motor":2, "Motor":3}
    name = None
    file = None #mesh file used by this body
    types = {"FrontLeft":[0,False], "FrontRight":[0,True], "RearLeft":[1,False], "RearRight":[1,True]}
    __axisTypes = {"Front":0, "Rear":1, "FrontLeft":0, "FrontRight":1, "RearLeft":2, "RearRight":3}
    __motionBehavior = {"Connect":0}
    __dynamic = None
    __material = {} #this will store the texture information in a dictinoary having part:materialname relationship
    __paths = {}    #all the added paths    
    __mesh = None
    
    def __init__ (self, name) :
        """
        Creates a physical vehicle having name 'name' without wheels. The chassis is a default object and is 
        changed by calling setChassis(...).
        """
        self.id = name        
        _setVehicleProperty(self.__PhysicsVehicleProperty_Create, {"Name":self.id})
        self.__mesh = None

    def attach(self, body, pos=[0,0,0], lforce=0, aforce=0) : 
        """ 
        Attaches the body to vehicles chassis position in local space where 
        - force is the strenght of the attachement. Force must be positive or 0 value
        - pos is the position and 
        - body is the body instance
        """
        if lforce < 0 : raise VehicleError, "lforce was smaller than 0"
        if aforce < 0 : raise VehicleError, "aforce was smaller than 0"
        if len(pos) != 3 : raise VehicleError, "The position must be a array having 3 real values"
        _setVehicleProperty (self.__PhysicsVehicleProperty_Attach, {"Name":self.id,"BodyTarget":body.id,"LinearForce": lforce, "AngularForce": aforce, "PositionX":pos[0],"PositionY":pos[1],"PositionZ":pos[2]})
    
    def getName(self) : return self.id

    def makeDynamic (self) :
        """
        Makes the actual vehicle according to the specified settings set through the set calls.
        The representation is dynamic meaning it will be drivable and behave as an vehicle depending
        on the settings.
        """
        _setVehicleProperty(self.__PhysicsVehicleProperty_MakeDynamic, {"Name":self.id})
        self.__dynamic = True
        self.__mesh = self.getPartName("chassi")[0]

    def makeStatic (self) :
        """
        Makes the an static vehicle according to the specified settings set through the set calls.
        The representation is static meaning it will NOT be drivable. 
        """
        _setVehicleProperty(self.__PhysicsVehicleProperty_MakeStatic, {"Name":self.id})
        self.__dynamic = False
        self.__mesh = self.getPartName("chassi")[0]
        
    def isDynamic (self) : return self.__dynamic

    def setPosition (self, x, y, z) :
        """ 
        Sets the vehicles position in global space. The y position will be calculated "ground height + y". If y is None we use the ground height for y.
        """
        print "Vehicle.setPostion ",x,y,z
        _setVehicleProperty (self.__PhysicsVehicleProperty_Position, {"Name":self.id,"PositionX":x,"PositionY":y,"PositionZ":z})

    def getPosition (self) :
        """ 
        Gets the vehicles position in global space.
        """
        p = eval(_getVehicleProperty (self.__PhysicsVehicleProperty_Position, {"Name":self.id}).replace(" ", ", "))
        return p

    def setOrientation (self, w, x, y, z) :
        """
        Sets the vehicles orientation of type quaternion in global space.
        """
        _setVehicleProperty (self.__PhysicsVehicleProperty_Orientation, {"Name":self.id,"RotationW":w,"RotationX":x,"RotationY":y,"RotationZ":z})

    def getOrientation (self) :
        """
        Returns a quaternion of the vehicles orientation
        """
        o = eval(_getVehicleProperty (self.__PhysicsVehicleProperty_Orientation, {"Name":self.id}).replace(" ", ", "))
        return o

    def getRotation (self) :
        """
        Returns a array having [x,y,z] degree rotations
        __MovableObjectProperty_Yaw = 4
        __MovableObjectProperty_Roll = 5
        __MovableObjectProperty_Pitch = 6
        
        """
        if self.__mesh <> None: 
            dic = {"Name":self.__mesh}
            x = gameGraphics.MovableObject_getProperty ( 6, str( dic ) )
            y = gameGraphics.MovableObject_getProperty ( 4, str( dic ) )
            z = gameGraphics.MovableObject_getProperty ( 5, str( dic ) )
            return [x,y,z]

    def setChassis(self, mesh, mass, massoffset=[0,-1,0]):
        """
        Sets the vehicles chassis to be mesh having mass and an massoffset. 
        Where mass must be a positive value. The mesh must exist in resources.
        """
        if mass <= 0 : raise VehicleError, "Setting vehicle chassis failed because mass is none positive, " + str(mass)
        if mesh == "" : raise VehicleError, "Setting vehicle chassis failed because you did not set the mesh name. It is NULL string"
        if len(massoffset) <> 3 : raise VehicleError, "The mass position defined by massposition must have 3 float values, got " + str(massoffset)
        dic = {"Name":self.id,"MeshFile":mesh,"Mass":mass}
        dic["MassOffsetX"] = massoffset[0]
        dic["MassOffsetY"] = massoffset[1]
        dic["MassOffsetZ"] = massoffset[2]
        _setVehicleProperty (self.__PhysicsVehicleProperty_Chassis, dic)
        
    def setWheelTracks (self, materialName) :
        """
        Sets vehicles wheel track material. The material named 'materialName' must be loaded by Ogre. 
        The material named 'materialName' will imidiatly be used on current tracks. 
        If null then wheel tracks are disabled.
        """
        _setVehicleProperty (self.__PhysicsVehicleProperty_WheelTracks, {"Name":self.id, "TrackMaterialName":materialName} )
        
    def setWheelParticles(self, particleMaterial="", channel=0, delta=0) :
        """
        Sets the vehicles wheel particle material. The material must be loaded or loadable by Ogre. 
        The material will be set to the wheel imidiatly and used when driving depending on settings provided. 
        Where particleMaterial is the name of an existing accessable ogre material
        and channel is the channel to be assigned the particles (channels are integer values from 0 to 3)
        and delta is the delta value to be added where we start our emission from (for instance if rotation 
        power is .3 and this is .5 then output will be .8)
        If particleMaterial is NULL ('') then the particles for channel 'channel' is removed.
        Raises VehicleError on failure.
        """
        if channel < 0 or channel > 3: raise VehicleError, "channel must be values between 0 and 3. Got " +str (channel)
        dic = {"Name":self.id, "ParticleMaterialName":particleMaterial, "ParticleChannel":channel, "ParticleEmissionDelta":delta}
        _setVehicleProperty (self.__PhysicsVehicleProperty_WheelParticles, dic)

    def setTireGrip (self, axis, longitudal, lateral):
        """
        Sets the vehicle tires grip on 'axis' where axis is either 'Front' or 'Rear'. 
        The grip is the stiffness factor dependent on force.
        Must have positive values. Raises VehicleError on failure.
        """
        types = {"Front":0, "Rear":1}
        if not types.has_key (axis) : raise VehicleError, "Adding vehicle axis failed because no such axis exists '" + str(axis) + "'. Supported are " + str(types.keys())        
        if longitudal < 0 : raise VehicleError, "Could not set grip because longitudal was negative, got " + str(longitudal)
        if lateral < 0 : raise VehicleError, "Could not set grip because lateral was negative, got " + str(lateral)
        _setVehicleProperty (self.__PhysicsVehicleProperty_TireGrip, {"Name":self.id, "AxisType":types[axis],"GripLongitudinal":longitudal,"GripLateral":lateral})

#    def setTireMass (self, mass, mass_inverted):
#        """
#        Sets the vehicle tires mass (on all four). Must have positive values. 
#        Raises VehicleError on failure.
#        """
#        if mass < 0 : raise VehicleError, "Could not set mass because it was negative, got " + str(mass)
#        if mass_inverted < 0 : raise VehicleError, "Could not set mass because mass_inverted was negative, got " + str(mass_inverted)
#        _setVehicleProperty (self.__PhysicsVehicleProperty_TireMass, {"Name":self.id,"Mass":mass,"MassInverted":mass_inverted})

    def setTireBehavior (self, axis, LateralAsymptoteSlip, LateralAsymptoteValue, LateralExtremumSlip, LateralExtremumValue, LongitudalAsymptoteSlip, LongitudalAsymptoteValue, LongitudalExtremumSlip, LongitudalExtremumValue):
        """
        The behavior of the tire is dependent on these values. They are used to calculate the grip with relation to force and wheel rotation.
        See PhysX documentation on NxWheelShape for detailed information.
        'axis' is either 'Front' or 'Rear'. 
        LateralAsymptoteSlip, LateralAsymptoteValue, LateralExtremumSlip, LateralExtremumValue.
        LongitudalAsymptoteSlip, LongitudalAsymptoteValue, LongitudalExtremumSlip, LongitudalExtremumValue.
        Must be positive values, LateralExtremumSlip < LateralAsymptoteSlip and LongitudalExtremumSlip < LongitudalAsymptoteSlip
        Raises VehicleError on failure.
        """
        types = {"Front":0, "Rear":1}
        if not types.has_key (axis) : raise VehicleError, "Adding vehicle axis failed because no such axis exists '" + str(axis) + "'. Supported are " + str(types.keys())        
        if LateralExtremumSlip > LateralAsymptoteSlip : raise VehicleError, "Could not set tire behvoir because LateralExtremumSlip > LateralAsymptoteSlip" 
        if LongitudalExtremumSlip > LongitudalAsymptoteSlip : raise VehicleError, "Could not set tire behvoir because LongitudalExtremumSlip > LongitudalAsymptoteSlip" 

        if LateralAsymptoteSlip < 0: raise VehicleError, "Could not set tire behvoir because we got an negative value" 
        if LateralAsymptoteValue < 0: raise VehicleError, "Could not set tire behvoir because we got an negative value" 
        if LateralExtremumSlip < 0: raise VehicleError, "Could not set tire behvoir because we got an negative value" 
        if LateralExtremumValue < 0: raise VehicleError, "Could not set tire behvoir because we got an negative value" 
        if LongitudalAsymptoteSlip < 0: raise VehicleError, "Could not set tire behvoir because we got an negative value" 
        if LongitudalAsymptoteValue < 0: raise VehicleError, "Could not set tire behvoir because we got an negative value" 
        if LongitudalExtremumSlip < 0: raise VehicleError, "Could not set tire behvoir because we got an negative value" 
        if LongitudalExtremumValue < 0: raise VehicleError, "Could not set tire behvoir because we got an negative value" 
        dic = {"Name":self.id, "AxisType":types[axis]}
        dic["LateralAsymptoteSlip"] = LateralAsymptoteSlip
        dic["LateralAsymptoteValue"] = LateralAsymptoteValue
        dic["LateralExtremumSlip"] = LateralExtremumSlip
        dic["LateralExtremumValue"] = LateralExtremumValue
        
        dic["LongitudalAsymptoteSlip"] = LongitudalAsymptoteSlip
        dic["LongitudalAsymptoteValue"] = LongitudalAsymptoteValue
        dic["LongitudalExtremumSlip"] = LongitudalExtremumSlip
        dic["LongitudalExtremumValue"] = LongitudalExtremumValue
        
        _setVehicleProperty (self.__PhysicsVehicleProperty_TireBehavior, dic)

    def setSteer(self, SteerServoIncrease=4, SteerServoDecrease=4, SteerMax=0.6, RotateOwnAxisOnStop=0):
        """
        Sets the servo steering of the wheels. Where SteerServoIncrease is the speed in which we turn the wheel and SteerServoDecrease is the
        speed the wheels restores themselfs when player stops steering. SteerMax is the maximum angle (in radians) that we may steer at either direction.
        The speed in which we turn must be not null positive values and SteerMax maximum value is 1.4 (80 degrees).
        Raises VehicleError on failure.
        """
        if SteerServoIncrease < 0 : raise VehicleError, "Could not set steering properties, got negative value"
        if SteerServoDecrease < 0 : raise VehicleError, "Could not set steering properties, got negative value"
        if SteerMax < 0 : raise VehicleError, "Could not set steering properties, got negative value"
        if SteerMax > 1.4 : raise VehicleError, "Could not set steering properties because SteerMax is larger than 1.4 radians, got " + str(SteerMax)
        _setVehicleProperty (self.__PhysicsVehicleProperty_Steer, {"Name":self.id,"SteerServoIncrease":SteerServoIncrease,"SteerServoDecrease":SteerServoDecrease,"SteerMax":SteerMax,"SteerFrontRear":RotateOwnAxisOnStop})

    def setBrake(self, axis, force=100):
        """
        Sets the vehicle brake force. May not be negative null value and axis is either 'Front' or 'Rear'
        Raises VehicleError on failure.
        """
        types = {"Front":0, "Rear":1}
        if not types.has_key (axis) : raise VehicleError, "Setting vehicle brake failed because no such axis exists '" + str(axis) + "'. Supported are " + str(types.keys())                
        if force < 0 : raise VehicleError, "Could not set brake properties, got negative value"
        _setVehicleProperty (self.__PhysicsVehicleProperty_Brake, {"Name":self.id,"Force":force, "AxisType":types[axis]})
    
    def addAxis (self, type, meshName, zposition):
        """
        Will add an axis of type 'type' having mesh 'meshName'. The Mesh must be available
        to the ogre resources.
        """
        types = {"Front":0, "Rear":1}
        if not types.has_key (type) : raise VehicleError, "Adding vehicle axis failed because no such axis exists '" + str(type) + "'. Supported are " + str(types.keys())
        if meshName == "" : raise VehicleError, "Adding vehicle axis failed because meshName was NULL"
        dic = {"Name":self.id}
        dic["MeshName"] = meshName
        dic["AxisType"] = types[type]
        dic["ZPosition"] = zposition
        _setVehicleProperty (self.__PhysicsVehicleProperty_AddAxis, dic)

    def addWheel (self, type, meshName):
        """
        Raises VehicleError on failure.
        """
        dic = {"Name":self.id}
        if meshName == "" : raise VehicleError, "Adding vehicle wheel failed because you did not set the mesh name. It is null string"
        if not self.types.has_key (type) : raise VehicleError, "Adding vehicle wheel failed because no such type exists '" + str(type) + "'. Supported are " + str(self.types.keys())
        dic["MeshName"] = meshName
        dic["RightSide"] = str(self.types[type][1]).lower()
        dic["AxisType"] = self.types[type][0]
        _setVehicleProperty (self.__PhysicsVehicleProperty_AddWheel, dic)

    def delWheel (self, type) :
        """
        Removes the wheel from current vehicle setup. Raises VehicleError if the wheel does not exist.
        Raises VehicleError if the wheel does not exist.
        """
        if not self.types.has_key (type) : raise VehicleError, "Adding vehicle wheel failed because no such type exists '" + str(type) + "'. Supported are " + str(self.types.keys())
        dic = {"Name":self.id}
        dic["RightSide"] = str(self.types[type][1]).lower()
        dic["AxisType"] = self.types[type][0]
        _setVehicleProperty (self.__PhysicsVehicleProperty_DelWheel, dic)

    def setEngine (self, power, increase, decrease) :
        """
        Sets the vehicles engine properties where power is the maximum power through output and increase / decrease is the speed in with the
        engine increase and decrease its output. These are values must be larger than 0 (good values are 60, 0.8, 1). They decide how fast the engine
        will go from 0% to 100% during gas intake and gas halt. Where at 100% it will give defined power.
        Call addPowerStep to tune how much the engine will output at these steps.
        Raises VehicleError on failure.
        """
        if power <= 0 : raise VehicleError, "Could not setup engine attributes because 'power' was 0 or negative value"
        if increase <= 0 : raise VehicleError, "Could not setup engine attributes because 'increase' was 0 or negative value"
        if decrease <= 0 : raise VehicleError, "Could not setup engine attributes because 'decrease' was 0 or negative value"
        _setVehicleProperty (self.__PhysicsVehicleProperty_EngineSetup, {"Name":self.id, "EngineIncrease":increase, "EngineDecrease":decrease, "EnginePower":power})

    def addEnginePowerStep (self, ratio) :
        """
        Adds a power step to the engine so that it will output value ratio*power at this step. The step position is set
        with an internal call that increases and decreases with values set by calling setEngine(...).
        This value must be 0 to 1 where 1 will be the maximum value set. There can be many steps taken for an engine to
        reach 1. The first value should be 0 while the last is 1.
        You can clear all steps by calling delEnginePowerSteps(...). 
        Raises VehicleError on failure.
        """
        if ratio < 0 : raise VehicleError, "Could not set engine power step since it was lower than 0"
        if ratio > 1 : raise VehicleError, "Could not set engine power step since it was higher than 1"
        _setVehicleProperty (self.__PhysicsVehicleProperty_EnginePowerStep, {"Name":self.id, "Step":ratio});

    def delEnginePowerSteps (self) :
        """
        Clears engine steps added during call of addEnginePowerStep(...). Will revert to linear increase of power.
        """
        _setVehicleProperty (self.__PhysicsVehicleProperty_EnginePowerStepClear, {"Name":self.id})

    def setCorrection (self, posrot) :
        """
        Corrects the vehicle position and orientation if needed where posrot is a string having 
        format [[x, y ,z],[w, x, y, z]]. This string is turned into 
        two arrays. First having 3 values (x,y,z) and the second having quaternion (w,x,y,z). 
        Returns true if correction was made false if no correction was needed.
        """
        posrot = eval(posrot)
        pos = posrot[0]
        rot = posrot[1]
        dic = {"Name":self.id}
        dic["px"] = pos[0]
        dic["py"] = pos[1]
        dic["pz"] = pos[2]
        dic["ow"] = rot[0]
        dic["ox"] = rot[1]
        dic["oy"] = rot[2]
        dic["oz"] = rot[3]
        _getVehicleProperty (self.__PhysicsVehicleProperty_Correction, dic)

    def getCorrection (self) :
        """
        Gets this vehicles correct possition and rotation. This is used to update the network simulated
        vehicles. Returns [[x, y ,z],[w, x, y, z]] formated string.
        """
        return str(_getVehicleProperty (self.__PhysicsVehicleProperty_PositionOrientationExpected, {"Name":self.id})).replace(" ",", ")
   
    def setDeadreconing (self, data) :
       _setVehicleProperty (self.__PhysicsVehicleProperty_DeadRecon, {"Name":self.id, "Data":data})

    def getDeadreconing (self) :
       return _getVehicleProperty (self.__PhysicsVehicleProperty_DeadRecon, {"Name":self.id})

    def getPartName (self, part) :
        """
        Gets this vehicles generated name of the parts entity. The Parameter part must be one of 
        these supported strings: 'chassi' and 'tire'
        """
        parttypes = ['chassi', 'tire']
        if not part in parttypes : raise VehicleError, "Could not get part name since '"+part+"' is not supported. Supported types are " + str(parttypes)
        return eval(_getVehicleProperty (self.__PhysicsVehicleProperty_Name, {"Name":self.id, "Part":part}))

    def useMaterial (self) :
        """
        Will use the previus assigned material on all the objects. This is mainly so we do not need to 
        get the information again (we just reuse it).
        Raises VehicleError if dictionary was not set at all (no succesfull call to setMaterial was made)
        """
        if len (self.__material) == 0: raise VehicleError, "Trying to reuse material but no material was set through call setMaterial()"
        for part in self.__material : self.setMaterial (part, self.__material[part])
        
    def setMaterial (self, part, material) :
        """
        Sets this vehicles generated entities to te specified material 'material' of the vehicle. 
        The parameter 'part' must be one of these supported strings: see getPartName.
        Raises VehicleError on failure. Stores the material on success and can be reassigned through call useMaterial()
        """
        entities = []
        names = self.getPartName(part)
        if len(names) == 0 : raise VehicleError, "Could not set parts of type '" + part + "' to use material '"+material+"' because the get part name returned []"
        for name in names : 
            if gameGraphics.Entity_setProperty ( 1, str( {"Name":name, "MaterialName":material} ) ) == 0 :
                raise VehicleError, "Fail to set parts because of entity object set property error" 
        self.__material[part] = material

    def getMaterial (self, part) :
        """
        Returns the assinged material of the requested part. None if no material was defined for the part.
        The successfull return value is material name
        """
        if self.__material.has_key(part) :  return self.__material[part]

   
    def destroy (self) :
        """
        Removes the vehicle and current attached parts to the vehicle such as wheel, engine, suspension, chassis, etc ...
        """
        _setVehicleProperty (self.__PhysicsVehicleProperty_Destroy, {"Name":self.id})

    def setDriveShaftMotion (self, axisType, boltType, meshName, pivotOffset=[0,0,0], behavior="Connect"):
        """
        Creates an visual representation of the drive shaft (can be omitted). Where 
        axisType is either 'FrontLeft', 'FrontRight', 'RearLeft', 'RearRight'
        boltType is created through calls setChassisBolt and setAxisBolt
        meshName if name of file that is accessable by ogre resource system.
        pivotSystem has three float values positioning the pivot of the vehicle part (where pivot is the axis which we rotate around)
        behavoir is either 'Connect' ... TBD
        If this call is not called then no visual representation will be created for this vehicle part but it will
        function properly.
        Raises VehicleError on failure
        """
        if not self.__axisTypes.has_key(axisType) : raise VehicleError, "Could not define vehicle motion because axisType='"+axisType+"' is not supported, supported types " + str(self.__axisTypes)
        if boltType == "" : raise VehicleError, "Could not define vehicle motion because boltType is NULL"
        if meshName == "" : raise VehicleError, "Could not define vehicle motion because meshName is NULL"
        if len(pivotOffset) <> 3 : raise VehicleError, "Could not define vehicle motion because pivotOffset must have three float values got, " + str(pivotOffset)
        if not self.__motionBehavior.has_key(behavior) : raise VehicleError, "Could not define vehicle motion because axisType='"+behavior+"' is not supported, supported types " + str(self.__motionBehavior)
        dic = {"Name":self.id}
        dic["BoltType"] = boltType
        dic["AxisType"] = self.__axisTypes[axisType]
        dic["MeshFile"] = meshName
        dic["Behavior"] = self.__motionBehavior[behavior]
        dic["PivotOffsetX"] = pivotOffset[0]
        dic["PivotOffsetY"] = pivotOffset[1]
        dic["PivotOffsetZ"] = pivotOffset[2]
        _setVehicleProperty (self.__PhysicsVehicleProperty_DriveShaft, dic)


    def setSuspensionMotion (self, axisType, boltType, meshName, pivotOffset=[0,0,0], behavior="Connect"):
        """
        Creates an visual representation of the suspenion (can be omitted). Where 
        axisType is either 'FrontLeft', 'FrontRight', 'RearLeft', 'RearRight'
        boltType is created through calls setChassisBolt and setAxisBolt
        meshName if name of file that is accessable by ogre resource system.
        pivotSystem has three float values positioning the pivot of the vehicle part (where pivot is the axis which we rotate around)
        behavoir is either 'Connect' ... TBD
        If this call is not called then no visual representation will be created for this vehicle part but it will
        function properly.
        Raises VehicleError on failure
        """
        if not self.__axisTypes.has_key(axisType) : raise VehicleError, "Could not define vehicle motion because axisType='"+axisType+"' is not supported, supported types " + str(self.__axisTypes)
        if boltType == "" : raise VehicleError, "Could not define vehicle motion because boltType is NULL"
        if meshName == "" : raise VehicleError, "Could not define vehicle motion because meshName is NULL"
        if len(pivotOffset) <> 3 : raise VehicleError, "Could not define vehicle motion because pivotOffset must have three float values got, " + str(pivotOffset)
        if not self.__motionBehavior.has_key(behavior) : raise VehicleError, "Could not define vehicle motion because axisType='"+behavior+"' is not supported, supported types " + str(self.__motionBehavior)
        dic = {"Name":self.id}
        dic["BoltType"] = boltType
        dic["AxisType"] = self.__axisTypes[axisType]
        dic["MeshFile"] = meshName
        dic["Behavior"] = self.__motionBehavior[behavior]
        dic["PivotOffsetX"] = pivotOffset[0]
        dic["PivotOffsetY"] = pivotOffset[1]
        dic["PivotOffsetZ"] = pivotOffset[2]
        _setVehicleProperty (self.__PhysicsVehicleProperty_Suspension, dic)

    def setSuspensionBehavior (self, axisType, suspensionSpring) :
        """
        Sets the control arm behavoir where 
        axisType is either 'Front' or 'Rear' and
        suspensionSpring is larger than zero.
        Raises VehicleError on failure
        """
        if not self.__axisTypes.has_key(axisType) : raise VehicleError, "Could not define vehicle behavoir because axisType='"+axisType+"' is not supported, supported types " + str(self.__axisTypes)
        if suspensionSpring <= 0 : raise VehicleError, "Could not define vehicle motion because controlArmTravel is 0 or lower value"
        dic = {"Name":self.id}
        dic["AxisType"] = self.__axisTypes[axisType]
        dic["SuspensionSpring"] = suspensionSpring
        _setVehicleProperty (self.__PhysicsVehicleProperty_Suspension, dic)

    def setShockAbsorberMotion (self, axisType, boltType, meshName, pivotOffset=[0,0,0], behavior="Connect"):
        """
        Creates an visual representation of the shock absorber (can be omitted). Where 
        axisType is either 'FrontLeft', 'FrontRight', 'RearLeft', 'RearRight'
        boltType is created through calls setChassisBolt and setAxisBolt
        meshName if name of file that is accessable by ogre resource system.
        pivotSystem has three float values positioning the pivot of the vehicle part (where pivot is the axis which we rotate around)
        behavoir is either 'Connect' ... TBD
        If this call is not called then no visual representation will be created for this vehicle part but it will
        function properly.
        Raises VehicleError on failure
        """
        if not self.__axisTypes.has_key(axisType) : raise VehicleError, "Could not define vehicle motion because axisType='"+axisType+"' is not supported, supported types " + str(self.__axisTypes)
        if boltType == "" : raise VehicleError, "Could not define vehicle motion because boltType is NULL"
        if meshName == "" : raise VehicleError, "Could not define vehicle motion because meshName is NULL"
        if len(pivotOffset) <> 3 : raise VehicleError, "Could not define vehicle motion because pivotOffset must have three float values got, " + str(pivotOffset)
        if not self.__motionBehavior.has_key(behavior) : raise VehicleError, "Could not define vehicle motion because axisType='"+behavior+"' is not supported, supported types " + str(self.__motionBehavior)
        dic = {"Name":self.id}
        dic["BoltType"] = boltType
        dic["AxisType"] = self.__axisTypes[axisType]
        dic["MeshFile"] = meshName
        dic["Behavior"] = self.__motionBehavior[behavior]
        dic["PivotOffsetX"] = pivotOffset[0]
        dic["PivotOffsetY"] = pivotOffset[1]
        dic["PivotOffsetZ"] = pivotOffset[2]
        _setVehicleProperty (self.__PhysicsVehicleProperty_ShockAbsorbers, dic)
    
    def setShockAbsorberBehavior (self, axisType, shockAbsorberDamper, shockAbsorberTarget) :
        """
        Sets the control arm behavoir where 
        axisType is either 'FrontLeft', 'FrontRight', 'RearLeft', 'RearRight' and
        shockAbsorberDamper is larger than zero and
        shockAbsorberTarget is between zero and one.
        Raises VehicleError on failure
        """
        if not self.__axisTypes.has_key(axisType) : raise VehicleError, "Could not define vehicle behavoir because axisType='"+axisType+"' is not supported, supported types " + str(self.__axisTypes)
        if shockAbsorberDamper <= 0 : raise VehicleError, "Could not define vehicle motion because controlArmTravel is 0 or lower value"
        if not validposreal(shockAbsorberTarget) : raise VehicleError, "Could not define vehicle motion because controlArmTravel must be a value between zero and one. Got " + str(shockAbsorberTarget)
        dic = {"Name":self.id}
        dic["AxisType"] = self.__axisTypes[axisType]
        dic["ShockAbsorberDamper"] = shockAbsorberDamper
        dic["ShockAbsorberTarget"] = shockAbsorberTarget
        _setVehicleProperty (self.__PhysicsVehicleProperty_ShockAbsorbers, dic)
        
    def setControlArmMotion (self, axisType, boltType, meshName, pivotOffset=[0,0,0], behavior="Connect"):
        """
        Creates an visual representation of the control arm (can be omitted). Where 
        axisType is either 'FrontLeft', 'FrontRight', 'RearLeft', 'RearRight'
        boltType is created through calls setChassisBolt and setAxisBolt
        meshName if name of file that is accessable by ogre resource system.
        pivotSystem has three float values positioning the pivot of the vehicle part (where pivot is the axis which we rotate around)
        behavoir is either 'Connect' ... TBD
        If this call is not called then no visual representation will be created for this vehicle part but it will
        function properly.
        Raises VehicleError on failure
        """
        if not self.__axisTypes.has_key(axisType) : raise VehicleError, "Could not define vehicle motion because axisType='"+axisType+"' is not supported, supported types " + str(self.__axisTypes)
        if boltType == "" : raise VehicleError, "Could not define vehicle motion because boltType is NULL"
        if meshName == "" : raise VehicleError, "Could not define vehicle motion because meshName is NULL"
        if len(pivotOffset) <> 3 : raise VehicleError, "Could not define vehicle motion because pivotOffset must have three float values got, " + str(pivotOffset)
        if not self.__motionBehavior.has_key(behavior) : raise VehicleError, "Could not define vehicle motion because axisType='"+behavior+"' is not supported, supported types " + str(self.__motionBehavior)
        dic = {"Name":self.id}
        dic["BoltType"] = boltType
        dic["AxisType"] = self.__axisTypes[axisType]
        dic["MeshFile"] = meshName
        dic["Behavior"] = self.__motionBehavior[behavior]
        dic["PivotOffsetX"] = pivotOffset[0]
        dic["PivotOffsetY"] = pivotOffset[1]
        dic["PivotOffsetZ"] = pivotOffset[2]
        _setVehicleProperty (self.__PhysicsVehicleProperty_ControlArms, dic)

    def setControlArmBehavior (self, axisType, controlArmTravel) :
        """
        Sets the control arm behavoir where 
        axisType is either 'FrontLeft', 'FrontRight', 'RearLeft', 'RearRight' and
        controlArmTravel is larger than zero.
        Raises VehicleError on failure
        """
        if not self.__axisTypes.has_key(axisType) : raise VehicleError, "Could not define vehicle behavoir because axisType='"+axisType+"' is not supported, supported types " + str(self.__axisTypes)
        if controlArmTravel <= 0 : raise VehicleError, "Could not define vehicle motion because controlArmTravel is 0 or lower value"
        dic = {"Name":self.id}
        dic["AxisType"] = self.__axisTypes[axisType]
        dic["ControlArmTravel"] = controlArmTravel
        _setVehicleProperty (self.__PhysicsVehicleProperty_ControlArms, dic)

        
    def setWheelBehavior (self, axisType, mass, massinv) :
        """
        Sets the wheel behavior through mass and mass inverted values. Where 
        axisType is either 'FrontLeft', 'FrontRight', 'RearLeft', 'RearRight' and
        mass and massinv is not negative values.
        """
        if not self.__axisTypes.has_key(axisType) : raise VehicleError, "Could not define vehicle motion because axisType='"+axisType+"' is not supported, supported types " + str(self.__axisTypes)
        if mass < 0 : raise VehicleError, "Could not define vehicle motion because mass was an negative value."
        if massinv < 0 : raise VehicleError, "Could not define vehicle motion because massinv was an negative value."
        dic = {"Name":self.id}
        dic["AxisType"] = self.__axisTypes[axisType]
        dic["WheelMass"] = mass
        dic["WheelMassInverted"] = massinv
        _setVehicleProperty (self.__PhysicsVehicleProperty_Wheel, dic)
        
    def setChassisBolt (self, boltName, meshName, offset=[0,0,0]) :
        """
        Creates and attaches an bolt to the chassis. The 'boltName' is used to connect another vehicle part to this bold.
        The bolt is sort of a market that is used to identify the location. 
        'meshName' is the visual representation. The meshName must be an resource available mesh or null. If null no mesh
        is used.
        'offset' is a three dimensional location relative to the parent 'chassis' center (or in other words, parent space).
        Raises VehicleError on failure
        """
        if boltName == "" : raise VehicleError, "Adding vehicle axis failed because boltName was NULL"        
        if len(offset) <> 3 : raise VehicleError, "The offset position must have 3 float values, got " + str(offset)
        dic = {"Name":self.id}
        dic["BoltName"] = boltName
        dic["MeshName"] = meshName
        dic["OffsetX"] = offset[0]
        dic["OffsetY"] = offset[1]
        dic["OffsetZ"] = offset[2]
        _setVehicleProperty (self.__PhysicsVehicleProperty_BoltFrame, dic)                
        
    def setAxisBolt (self, type, boltName, meshName, offset=[0,0,0]) :
        """
        Creates and attaches an bolt to the axis. The axis 'type' is either 'FrontLeft', 'FrontRight', 'RearLeft' or 'RearRight' denoting the location of the axis.
        The 'boltName' is used to connect another vehicle part to this bold.
        The bolt is sort of a market that is used to identify the location. 
        'meshName' is the visual representation. The meshName must be an resource available mesh or null. If null no mesh
        is used.
        'offset' is a three dimensional location relative to the parent 'chassis' center (or in other words, parent space).
        Raises VehicleError on failure
        """
        types = {"Front":0, "Rear":1, "FrontLeft":0, "FrontRight":1, "RearLeft":2, "RearRight":3}
        if not types.has_key (type) : raise VehicleError, "Adding vehicle axis failed because no such axis exists '" + str(type) + "'. Supported are " + str(types.keys())
        if boltName == "" : raise VehicleError, "Adding vehicle axis failed because boltName was NULL"        
        if len(offset) <> 3 : raise VehicleError, "The offset position must have 3 float values, got " + str(offset)
        dic = {"Name":self.id}
        dic["BoltName"] = boltName
        dic["MeshName"] = meshName
        dic["AxisType"] = types[type]
        dic["OffsetX"] = offset[0]
        dic["OffsetY"] = offset[1]
        dic["OffsetZ"] = offset[2]
        _setVehicleProperty (self.__PhysicsVehicleProperty_BoltAxis, dic)                

def createWorld (HasFloor="", bHasGravity=True) :
    """
    HasFloor takes a file name that represent its area. If config file is '' (nothing) then we create a floor at height 0.
    Creates a physical world returning None. Raises PhysicsError on failure
    """
    dic = {"HasFloor":HasFloor, "HasGravity":str(bHasGravity).lower()}
    _setWorldProperty(__PhysicsWorldProperty_Setup, dic)

def destroyWorld () :
    """
    Destroys a physical world and all of its physical contents if it exists returning None. 
    Raises PhysicsError on failure
    """
    dic = {"Name":"-"}
    _setWorldProperty(__PhysicsWorldProperty_Destroy, dic)


def debugEnabled (b, ip="127.0.0.1") :
    dic = {"DebugEnabled": str(b).lower()}
    dic["RemoteDebugIP"] = ip
    _setWorldProperty(__PhysicsWorldProperty_Debug, dic)

def delBody (name) :
    """
    Destorys the body and removes from scene. An body is created through calls
    addBall(), addBox(), addMesh(), addCapsule(), addPrism()
    """
    if not __bodies.has_key(name) : raise PhysicsError, "Could remove body because '" +name+"' does not exist in body list " + str(__bodies.keys())
    _setBodyProperty(__PhysicsBodyProperty_Destroy, {"Name":name})
    del __bodies[name]

def addBall (name, meshfile, mass=10, position=[0,0,0], static=False, radius=None) :
    """
    Creates a ball having using meshfile as the representation.
    If mass is not set then it will use 10 as default.
    If radius is not set then it will use meshes radius as default radius
    If position is not set then we use default 0,0,0
    """
    dic = {"Name":name, "MeshFile":meshfile}
    if __bodies.has_key(name) : raise PhysicsError, "Could not add ball. '" +name+"' already exists"
    if mass <= 0 : raise PhysicsError, "Could not add ball. '" +name+"' using mesh '"+meshfile+"' because the mass must be a positive value, got " + str(mass)
    if radius <> None: 
        if radius <= 0 : raise PhysicsError, "Could not add ball. '" +name+"' using mesh '"+meshfile+"' because the radius must be a positive value, got " + str(radius)
    if not validpos3(position) :  raise PhysicsError, "Could not add box. '" +name+"' using mesh '"+meshfile+"' because the position must have a list with three numbers, got " + str(position)
    dic["PositionX"] = position[0]
    dic["PositionY"] = position[1]
    dic["PositionZ"] = position[2]
    dic["Mass"] = mass
    dic["Static"] = str(static).lower()
    _setBodyProperty(__PhysicsBodyProperty_Ball, dic)
    __bodies[name] = Body(dic)
    return __bodies[name]

def addBox (name, meshfile, mass=10, position=[0,0,0], static=False, size=None) :
    """
    Creates a ball having using meshfile as the representation.
    If mass is not set then it will use 10 as default.
    If size is not set then it will use meshes size as default size
    If position is not set then we use default 0,0,0
    If static is true then we create a static object else not.
    """
    dic = {"Name":name, "MeshFile":meshfile}
    if __bodies.has_key(name) : raise PhysicsError, "Could not add box. '" +name+"' already exists"
    if mass <= 0 : raise PhysicsError, "Could not add box. '" +name+"' using mesh '"+meshfile+"' because the mass must be a positive value, got " + str(mass)
    if size <> None: 
        if not validsize3(size) : raise PhysicsError, "Could not add box. '" +name+"' using mesh '"+meshfile+"' because the size must have positive [x,y,z] values"
        else : 
            dic["SizeX"] = size[0]
            dic["SizeY"] = size[1]
            dic["SizeZ"] = size[2]            
    if not validpos3(position) :  raise PhysicsError, "Could not add box. '" +name+"' using mesh '"+meshfile+"' because the position must have a list with three numbers"
    dic["PositionX"] = position[0]
    dic["PositionY"] = position[1]
    dic["PositionZ"] = position[2]
    dic["Mass"] = mass        
    dic["Static"] = str(static).lower()
    _setBodyProperty(__PhysicsBodyProperty_Box, dic)
    __bodies[name] = Body(dic)
    return __bodies[name]

def addCapsule (name, meshfile, mass=10, position=[0,0,0], static=False, dim=None) :
    """
    Creates a capsule having using meshfile as the representation.
    If mass is not set then it will use 10 as default.
    If position is not set then we use default 0,0,0
    If static is True then we create an static mesh else not.
    If dim is set then we do not use the mesh dimension but the array of values that 
    must have 2 values [x,y]
    """
    dic = {"Name":name, "MeshFile":meshfile}
    if __bodies.has_key(name) : raise PhysicsError, "Could not add capsule. '" +name+"' already exists"
    if dim <> None : 
        if len(dim) <> 2 : raise PhysicsError, "Could not add capsule. '" +name+"' because the dim must be array of two values [x,y]"
        if not validsize3([dim[0],dim[1],1]) :  raise PhysicsError, "Could not add capsule. '" +name+"' using mesh '"+meshfile+"' because the dimension must have a list with two possitive numbers, got " + str(dim)
        dic["SizeX"] = dim[0]
        dic["SizeY"] = dim[1]
    if mass <= 0 : raise PhysicsError, "Could not add capsule. '" +name+"' using mesh '"+meshfile+"' because the mass must be a positive value, got " + str(mass)
    if not validpos3(position) :  raise PhysicsError, "Could not add capsule. '" +name+"' using mesh '"+meshfile+"' because the position must have a list with three numbers, got " + str(position)
    dic["PositionX"] = position[0]
    dic["PositionY"] = position[1]
    dic["PositionZ"] = position[2]
    dic["Mass"] = mass
    dic["Static"] = str(static).lower()
    _setBodyProperty(__PhysicsBodyProperty_Capsule, dic)
    __bodies[name] = Body(dic)
    return __bodies[name]
    
def addMesh (name, meshfile, mass=10, position=[0,0,0], static=False) :
    """
    Creates a capsule having using meshfile as the representation.
    If mass is not set then it will use 10 as default.
    If position is not set then we use default 0,0,0
    If static is True then we create an static mesh else not.
    Note that the mesh must be convex and follow PhysX requirements.
    """
    dic = {"Name":name, "MeshFile":meshfile}
    if __bodies.has_key(name) : raise PhysicsError, "Could not add ball. '" +name+"' already exists"
    if mass <= 0 : raise PhysicsError, "Could not add ball. '" +name+"' using mesh '"+meshfile+"' because the mass must be a positive value, got " + str(mass)
    if not validpos3(position) :  raise PhysicsError, "Could not add box. '" +name+"' using mesh '"+meshfile+"' because the position must have a list with three numbers, got " + str(position)
    dic["PositionX"] = position[0]
    dic["PositionY"] = position[1]
    dic["PositionZ"] = position[2]
    dic["Mass"] = mass
    dic["Static"] = str(static).lower()
    _setBodyProperty(__PhysicsBodyProperty_Convex, dic)
    __bodies[name] = Body(dic)
    return __bodies[name]    
    
def addPrism (name, meshfile, mass=10, position=[0,0,0], dim=[1,1,5], static=False) :
    """
    Creates a capsule having using meshfile as the representation.
    If mass is not set then it will use 10 as default.
    If position is not set then we use default 0,0,0
    If static is True then we create an static mesh else not.
    Dim must be an array of three values defining the width, height and side count
    """
    dic = {"Name":name, "MeshFile":meshfile}
    if __bodies.has_key(name) : raise PhysicsError, "Could not add ball. '" +name+"' already exists"
    if mass <= 0 : raise PhysicsError, "Could not add ball. '" +name+"' using mesh '"+meshfile+"' because the mass must be a positive value, got " + str(mass)
    if not validpos3(position) :  raise PhysicsError, "Could not add box. '" +name+"' using mesh '"+meshfile+"' because the position must have a list with three numbers, got " + str(position)
    if not validpos3(dim) : raise PhysicsError, "Could not add box. '" +name+"' using mesh '"+meshfile+"' because dimensions where not valid, see help"
    if dim[2] < 3 : raise PhysicsError, "Could not add prism. '" +name+"' using mesh '"+meshfile+"' because the prism must have more than 2 sides, got " + str(dim)
    dic["PositionX"] = position[0]
    dic["PositionY"] = position[1]
    dic["PositionZ"] = position[2]
    dic["SizeX"] = dim[0]
    dic["SizeY"] = dim[1]
    dic["SizeZ"] = dim[2]    
    dic["Mass"] = mass
    dic["Static"] = str(static).lower()
    _setBodyProperty(__PhysicsBodyProperty_Prism, dic)
    __bodies[name] = Body(dic)
    return __bodies[name]       
    
def getBody (name) : 
    if not __bodies.has_key(name) : return None
    else: return __bodies[name]
    
def addCar (name) :
    """
    Creates a vehicle instance and returns it to the caller.
    """
    __vehicles[name] = Vehicle(name)
    return __vehicles[name]

def getCar (name) :
    """
    Returns the car having id 'name' if it exists. Else returns None.
    """
    if __vehicles.has_key(name) :  return __vehicles[name]

def getCarNames () : 
    """
    Return all the loaded vehicle names
    """
    return __vehicles.keys()

def setCar (name, dic, deadreconing=None):
    dic["Name"] = name
    _setVehicleProperty(__PhysicsVehicleProperty_Input, dic)
    #if correction <> None : __vehicles[name].setCorrection(correction)
    if deadreconing <> None : __vehicles[name].setDeadreconing(deadreconing)

def delCar (name, dic):
    del __vehicles[name]

def isCarPresent (name) : return __vehicles.has_key(name)

def addPath (name, spline, materialName, deltaY=0.3, width=20, orientation=[1,0,0,0]) :
    """
    Adds a bezier path with name 'name' having zero points.
    The path sides will be as close to the ground as possible.
    Raises SceneError on failure.
    Returns None on success.
    """
    __PathProperty_Create = 0
    __PathProperty_AddPoint = 1
    __PathProperty_Flush = 2

    if len(spline) == 0 : raise SceneError, "Could not create track because there is no spline data"
    for t in spline : 
        if len(t) <> 2 :  raise SceneError, "Could not create track because an item in the list had more or less then two values, got " + str (spline)
    
    dic = {"Name":name, "MaterialName":materialName}
    dic["DeltaY"] = deltaY
    dic["PathWidth"] = width
    dic["OrientationW"] = orientation[0]
    dic["OrientationX"] = orientation[1]
    dic["OrientationY"] = orientation[2]
    dic["OrientationZ"] = orientation[3]    
    
    error = gamePhysics.Path_setProperty ( __PathProperty_Create, str(dic) )
    if  error <> None : raise SceneError, "SceneManager could not add '" + name + "' got error: " + str(error)
    for t in spline : 
        error = gamePhysics.Path_setProperty ( __PathProperty_AddPoint, str({"Name":name, "PointX":t[0], "PointZ":t[1]}) )
        if  error <> None : raise SceneError, "SceneManager could not create spline'" + name + "' got error: " + str(error)
    error = gamePhysics.Path_setProperty ( __PathProperty_Flush, str({"Name":name}) )
    if  error <> None : raise SceneError, "SceneManager could not generate '" + name + "' got error: " + str(error)
    __paths[name] = {"spline":spline, "material":materialName}    
    return __paths[name]

def delPath (name) :
    """
    Destroys a bezier path with name 'name'
    Raises SceneError on failure.
    Returns None on success.
    """
    __PathProperty_Destroy = 3
    if not __paths.has_key(name) : raise SceneError, "Could not delete bezier path '"+str(name)+"' because it does not exist"
    error = gamePhysics.Path_setProperty ( __PathProperty_Destroy, str({"Name":name}) )
    if  error <> None : raise SceneError, "SceneManager could not add '" + name + "' got error: " + str(error)
    del __paths[name]

def setAreaProperty (dic, x=None, y=None):
    """
    NOT SUPPORTED
    Sets an area property flusing the 'dic' into the system. The 'dic' attribute and values must be supported
    otherwise we raise SceneError exception.
    x and y is the page we assign the attributes onto. If omited we assign over the entire area.
    'dic' should use supported attributes and values. If we are describing something we should describe it fully.
    Currently we support 'define' of 'grass' and 'tree' where 
        'tree' input values are 'DetailLevelHigh' 'DetailLevelMedium', 'DetailLevelLow', 'DetailLevelCount', 'MeshName', 'RandomScaleMax' and 'RandomScaleMin'.
        'grass' input values are 'Density', 'DetailLevel', 'ColorMap', 'DensityMap', 'SizeMaxX','SizeMaxY','SizeMinX','SizeMinY','MaterialName','Radius','RandomScaleMax','RandomScaleMin','SwayDistribution','SwayLength','SwaySpeed'   
    """
    if (x <> None and y <> None) :  dic.update({"pageX":x, "pageY":y})
    _setWorldProperty (__PhysicsWorldProperty_Area, dic)

def setAreaRegion (descr):
    """
    descr is a xml file that we load with the region configuration.
    """
    _setWorldProperty (__PhysicsWorldProperty_Area, {"AreaRegion":descr})

def useMarker (meshName=None) :
    """
    A marker is an object that follows the ground in front of the camera (used for editing).
    Will enable/disable a terrain marker over a terrain. The marker is enabled if meshName is defined.
    Passing None or '' will disable the marker.
    Raises PhysicsError on failure.
    """
    if meshName == None : meshName = ""
    _setWorldProperty (__PhysicsWorldProperty_Marker, {"MarkerMesh":meshName})

def getMarker ():
    """
    A marker is an object that follows the ground in front of the camera (used for editing).
    A marker is created through call useMarker()
    Returns the position of the terrain marker
    """
    res = _getWorldProperty(__PhysicsWorldProperty_Marker, {"":None})
    res = eval(res.replace(" ", ", "))
    if res == None : raise PhysicsError, "No marker exists, create one with call to useMarker()"
    return res

def addGunTool (name, meshfile, force=1000, mass=10, distance=1000, position=[0,0,0]) :
    """
    Creates a gun tool having using meshfile as the representation.
    Where distance is the range of our gun. Objects outside the range is not effected
    If mass is not set then it will use 10 as default.
    If size is not set then it will use meshes size as default size
    If position is not set then we use default 0,0,0
    If static is true then we create a static object else not.
    """
    return addTool (__PhysicsToolProperty_Gun, name, meshfile, force, mass, distance, position)
    
def addRocketTool (name, meshfile, force=1000, mass=10, distance=1000, position=[0,0,0]) :
    """
    Creates a rocket tool having using meshfile as the representation.
    Where distance is the speed of the rocket.
    If mass is not set then it will use 10 as default.
    If size is not set then it will use meshes size as default size
    If position is not set then we use default 0,0,0
    If static is true then we create a static object else not.
    """
    return addTool (__PhysicsToolProperty_Rocket, name, meshfile, force, mass, distance, position)    
    
def addMagnetTool (name, meshfile, force=1000, mass=10, distance=1000, position=[0,0,0]) :
    """
    Creates a gun tool having using meshfile as the representation.
    Where distance is the range of our gun. Objects outside the range is not effected
    If mass is not set then it will use 10 as default.
    If size is not set then it will use meshes size as default size
    If position is not set then we use default 0,0,0
    If static is true then we create a static object else not.
    """
    return addTool (__PhysicsToolProperty_Magnet, name, meshfile, force, mass, distance, position)    

def addTool (tooltype, name, meshfile, force=1000, mass=10, distance=1000, position=[0,0,0]) :
    """
    Creates a tool having using meshfile as the representation.
    Where distance is the range of our gun. Objects outside the range is not effected
    If mass is not set then it will use 10 as default.
    If size is not set then it will use meshes size as default size
    If position is not set then we use default 0,0,0
    If static is true then we create a static object else not.
    """
    dic = {"Name":name, "MeshFile":meshfile}
    if __tools.has_key(name) : raise PhysicsError, "Could not add tool. '" +name+"' already exists"
    if mass <= 0 : raise PhysicsError, "Could not add box. '" +name+"' using mesh '"+meshfile+"' because the mass must be a positive value, got " + str(mass)
    dic["PositionX"] = position[0]
    dic["PositionY"] = position[1]
    dic["PositionZ"] = position[2]
    dic["Mass"] = mass        
    dic["Static"] = "false"
    dic["Range"] = distance
    dic["Force"] = force
    _setToolProperty(tooltype, dic)
    __tools[name] = Tool(dic)
    return __tools[name]

def delTool (name) :
    """
    Creates a tool having using meshfile as the representation.
    Where distance is the range of our gun. Objects outside the range is not effected
    If mass is not set then it will use 10 as default.
    If size is not set then it will use meshes size as default size
    If position is not set then we use default 0,0,0
    If static is true then we create a static object else not.
    """
    dic = {"Name":name}
    if not __tools.has_key(name) : raise PhysicsError, "Could not delete tool. '" +name+"' does not exists"
    _setToolProperty(__PhysicsToolProperty_Destroy, dic)
    del __tools[name]
	

def _setWorldProperty (attrib, dic) :
    """Private function used to set physical world properties. Raises PhysicsError on failure"""
    result = gamePhysics.World_setProperty (attrib, str(dic))
    if result <> None :    
        raise PhysicsError, "Could not set property " + str(attrib) + " using dictionary '" + str(dic) + "', faild setting state with supplied values, got: " + result

def _getWorldProperty (attrib, dic) :
    """Private function used to get physical world properties. Raises PhysicsError on failure"""
    res = gamePhysics.World_getProperty (attrib, str(dic))
    if res == None :
        raise PhysicsError, "Could not get property " + str(attrib) + " using dictionary '" + str(dic) + "', faild getting state with supplied values"
    else : return res


def _setBodyProperty (attrib, dic) :
    """Private function used to set physical body properties. Raises PhysicsError on failure"""
    result = gamePhysics.Body_setProperty (attrib, str(dic))
    if result <> None :
        raise PhysicsError, "Could not set property " + str(attrib) + " using dictionary '" + str(dic) + "', faild setting state with supplied values, got: " + result

def _getBodyProperty (attrib, dic) :
    """Private function used to get physical body properties. Raises PhysicsError on failure"""
    result = gamePhysics.Body_getProperty (attrib, str(dic))
    if result == None :
        raise PhysicsError, "Could not get property " + str(attrib) + " using dictionary '" + str(dic) + "', faild getting state with supplied values, got: " + result
    else : return result


def _setToolProperty (attrib, dic) :
    """Private function used to set physical tool properties. Raises PhysicsError on failure"""
    result = gamePhysics.Tool_setProperty (attrib, str(dic))
    if result <> None :
        raise ToolError, "Could not set property " + str(attrib) + " using dictionary '" + str(dic) + "', faild setting state with supplied values, got: " + result

def _getToolProperty (attrib, dic) :
    """Private function used to get physical tool properties. Raises PhysicsError on failure"""
    result = gamePhysics.Tool_getProperty (attrib, str(dic))
    if result == None :
        raise PhysicsError, "Could not get property " + str(attrib) + " using dictionary '" + str(dic) + "', faild getting state with supplied values, got: " + result
    else : return result


def _setVehicleProperty (attrib, dic) :
    """Private function used to set physical vehicle properties. Raises PhysicsError on failure"""    
    result = gamePhysics.Vehicle_setProperty (attrib, str(dic))
    if result <> None : 
        raise VehicleError, "Could not set property " + str(attrib) + " using dictionary '" + str(dic) + "', faild setting state with supplied values, got: " + result

def _getVehicleProperty (attrib, dic) :
    """Private function used to get physical vehicle properties. Raises PhysicsError on failure"""
    res = gamePhysics.Vehicle_getProperty (attrib, str(dic))
    if res == None :
        raise VehicleError, "Could not get property " + str(attrib) + " using dictionary '" + str(dic) + "', faild getting state with supplied values"
    else : return res

def _cleanUp () :
    for n in __vehicles.keys() :
        __vehicles[n].destroy()