"""This module behaves as a singleton allowing calls to the game graphics system. 
It will attempt to clean upp all the created objects before shutdown calling all
subsequent lighs, organics, entities, etc .. created to be deleted."""
__version__ = "$Revision: 1.64 $"
__source__ = "$Source: /cvsroot/game-client-system/scripts/SceneManager.py,v $"
__author__ = "Kjartan A Jonsson"
__copyright__ = "Copyright (c) 2007 aGameCompany"
__license__ = "aGameCompany"

from constants import *
from validatevalue import *
import gameGraphics

##############################################
#    TextureBlitter
#        Creates a texture that we can define areas where we blitt on in real time
##############################################
class TextureBlitterError (Exception) : pass
 
class TextureBlitter :
    __sources = None
    __targets = None
    name = None
    id = None
    width = None
    height = None 
    resourceGroup = None
    def __init__ (self, textureName, fileName, resourceGroup, width, height) :
        """
        Creates the instance and texture file. The textureName will be the name refered by material scripts and
        file name is the name of the file which we start from (this can be an empty image (64x64). The fileName must
        be accessable from the resource group 'resourceGroup'. The width and height must be larger than 0 and represents
        the final images size.
        This instance must be destroyed before exiting the application. Otherwise it causes a crash during unload.
        """        
        self.id = textureName
        self.width = width
        self.height = height
        self.resourceGroup = resourceGroup
        self.__sources = {}
        self.__targets = {}        
        error = gameGraphics.TextureBlitter_create (textureName, fileName, resourceGroup, width, height)
        if error <> None :
            raise TextureBlitterError, "Could not create texture blitter (with '"+textureName+"', '"+fileName+"', '"+str(width)+"', '"+str(height)+"'), got internal error:" + error
        
    def __checkDim (self, x, y, w, h):
        """
        Internal function to check the dimensions of the image. Raises TextureBlitterError on failure.
        """
        if not validposreal (x) : raise TextureBlitterError, "The supplied x value is not a value between 0 and 1, got " + str(x)
        if not validposreal (y) : raise TextureBlitterError, "The supplied y value is not a value between 0 and 1, got " + str(y)
        if not validposreal (w) : raise TextureBlitterError, "The supplied w value is not a value between 0 and 1, got " + str(w)
        if not validposreal (h) : raise TextureBlitterError, "The supplied h value is not a value between 0 and 1, got " + str(h)
        if not validposreal (x+w) : raise TextureBlitterError, "The supplied x+w value is not a value between 0 and 1 and exceeds canvas size, got " + str(x+w)
        if not validposreal (y+h) : raise TextureBlitterError, "The supplied y+h value is not a value between 0 and 1 and exceeds canvas size, got " + str(y+h)
        
    def addTarget (self, name, x, y, w, h) :
        """
        Adds a sticker location to the collection and sets its dimensions.
        Where targetName is the name to identify this location. x and y is the relative
        location on the bitmap (a number between 0 and 1) and w and h is the relative
        width and height of the target image.
        """
        self.__checkDim (x,y,w,h) #test for x,y,w,h
        if self.__targets.has_key(name) : raise TextureBlitterError, "The source name "+name+" already exists "
        error = gameGraphics.TextureBlitter_addDefinitionTarget (self.id, name, x, y, w, h)
        if error <> None : 
            raise TextureBlitterError, "Got error when applying target to internal blitter system. Got error: " + error
        self.__targets[name] = [x,y,w,h]

    def addSource (self, name, x, y, w, h, texture, resourceGroup="General", flipX=False, flipY=False) :    
        """
        Adds a sticker source definition to the collection and sets its dimensions.
        Where texture is the name of the texture file and x, y, w, h is the location within the
        texture where we take the image.
        If flipX is true then it will load this image and flip it in the x axis. 
        If flipY is true then it will load this image and flip it in the y axis.
        """        
        self.__checkDim (x,y,w,h) #test for x,y,w,h
        if self.__sources.has_key(name) : raise TextureBlitterError, "The source name "+name+" already exists " #test name exists
        #test texture not existant        
        iflipX = 0
        iflipY = 0
        if flipX : iflipX = 1
        if flipY : iflipY = 1
        error = gameGraphics.TextureBlitter_addDefinitionSource (self.id, name, texture, resourceGroup, x, y, w, h, iflipX, iflipY)   
        if error <> None :
            raise TextureBlitterError, "Got error when applying source '"+name+"' and texture '" + texture + "' to internal blitter system using data coordinates=" + str([x,y,w,h]) + ". Got error: "  + error   
        self.__sources[name] = [x,y,w,h]
        
    def paint (self, sourceName, targetName) :
        """
        Will set the sticker onto target location.
        """
        #test if sticker and target exists        
        gameGraphics.TextureBlitter_paint (self.id, sourceName, targetName)
        
    def __del__ (self) :
        """
        Destroys the texture blitter.
        """
        pass
#        error = gameGraphics.TextureBlitter_destroy (self.id)
#        if error <> None :
#            pass #raise TextureBlitterError, "Could not remove blitter '" + self.id + "' got error " + error
        
    def __str__ (self) :
        return "TextureBlitter name:"+self.id

class SceneError( Exception ): pass
class MovableObjectError ( SceneError ) : pass
class EntityError ( SceneError ) : pass
class LightError ( SceneError ) : pass
class OrganicError ( SceneError ) : pass
class ParticleError ( SceneError ) : pass
class CameraError ( SceneError ) : pass
class AnimationError ( SceneError ) : pass

class MovableObject :
    """Class representing a primitive object in the scene graph, is never used through direct initalization"""
    __MovableObjectProperty_Destroy = 0
    __MovableObjectProperty_Position = 1
    __MovableObjectProperty_Orientation = 2
    __MovableObjectProperty_Scale = 3
    __MovableObjectProperty_Yaw = 4
    __MovableObjectProperty_Roll = 5
    __MovableObjectProperty_Pitch = 6
    __MovableObjectProperty_LookAt = 7
    __MovableObjectProperty_Visible = 8
    __MovableObjectProperty_BoundingBox = 9    
    __MovableObjectProperty_Has = 10
    __MovableObjectProperty_Attach = 11
    __MovableObjectProperty_Size = 12
    __MovableObjectProperty_Distance = 13
    name = None
    __children = []
    
    def __init__( self ):
        """Abstract class initalization call that raises an error when called directly"""
        raise MovableObjectError, "Can not initiate MovableObject directly must use sub classe (Entity, Light, Organics, Particles)"

    def _exists( self ):
        """Checks to see if this object realy exists in system"""
        dic = {}
        dic["Name"] = self.id
        if gameGraphics.MovableObject_getProperty ( self.__MovableObjectProperty_Has, str( dic ) ) == "True" : return True
        else : raise MovableObjectError, "Object does not exist in system, add object using SceneManager"

    def setPosition ( self, x, y, z, type="Absolute", time = None, args = None ) :
        """
        Sets the position of our object, where type is either 'Absolute' or 'Relative'. Where 'Relative' uses the objects position as reference
        and sets the position relative to that. Absolute moves the object in worlds space. If time is set then we interpolate linearly from current 
        location to the desired (x, y, z) position during time 'time'. Where time is in milli seconds. 
        If arg is set then we interpolate using the  arg values as input to a spline (Ogre::SimpleSpline) function. The spline function takes two 
        list values each having three values that will be used to calculate the interpolation. The arg format is as follows [[x,y,z],[x,y,z]].
        Method raises MovableObjectError on failure and None on success.        
        """
        dic = {}
        dic["Name"] = self.id
        dic["x"] = x
        dic["y"] = y
        dic["z"] = z
        dic["Type"] = type
        if time <> None : 
            dic["Time"] = time
            if time < 0 : raise MovableObjectError, "Object could not oriented with interpolation because time was a negative number '"+str( time )+"'"
        if args <> None :
            if len( args ) == 2 : #the arguments must have a list with 2 arrays having 3 numbers
                if ( ( len( args[0] ) == 3 ) and ( len( args[1] ) == 3 ) ) : 
                   dic["sax"] = args[0][0]
                   dic["say"] = args[0][1]
                   dic["saz"] = args[0][2]
                   dic["sbx"] = args[1][0]
                   dic["sby"] = args[1][1]
                   dic["sbz"] = args[1][2]
                else : raise MovableObjectError, "Object could not position with spline interpolation because args must have two arrays each having three values got '"+str( args[0] )+"' and '"+str( args[1] )+"'"
            else : raise MovableObjectError, "Object could not position with spline interpolation because args must have two arrays each having three values got '"+str( args )+"'"
        self.__setProperty ( self.__MovableObjectProperty_Position, dic )
    def getPosition ( self, world=False ) :
        """Returns movable objects position in the form of a list having three float values [x, y, z] 
        If world is true then we will get the final position after all applied translation to parents. 
        If world is false then we get the translation in relation to parent."""
        result = []
        scale = self.__getProperty ( self.__MovableObjectProperty_Position, {"world":str(world).lower()})
        for i in scale.split( " " ) : result.append ( float( i ) )
        return result
    
    def getSize (self) :
        """
        Returns movable objects size in the form of a list having three float values [x, y, z] after scale
        (the bounding box size).
        """
        result = []
        size = self.__getProperty ( self.__MovableObjectProperty_Size, {})
        for i in size.split( " " ) : result.append ( float( i ) )
        return result
        
    def getDistanceTo (self, toObject):
        """
        Returns the distance to the object 'toObject'
        """
        pnt = self.__getProperty ( self.__MovableObjectProperty_Distance, {"Name":self.id, "MovableObject":toObject})
        return eval(pnt)
        
    def translate ( self, x, y, z, time=None, args=None ) :
        """Moves movable object relative to its parent"""
        self.setPosition ( x, y, z, "Relative", time, args )

    def setOrientation ( self, w, x, y, z, time=None, args=None ) :        
        """
        Sets the orientation of our object. If time is set then we interpolate linearly from current orientation to the desired (w, x, y, z) orientation 
        during time 'time'. Where time is in milli seconds. 
        If arg is set then we interpolate using the  arg values as input to a cubic (Ogre::Quaternion::Squad) function. The spline function takes two list 
        values each having four values that will be used to calculate the interpolation. The arg format is as follows [[w,x,y,z],[w,x,y,z]].
        Method raises MovableObjectError on failure and None on success.        
        """
        dic = {"Name":self.id, "w":w, "x":x, "y":y, "z":z}
        if time <> None : 
            dic["Time"] = time
            if time < 0 : raise MovableObjectError, "Object could not oriented with interpolation because time was a negative number '"+str( time )+"'"
        if args <> None :
            if len( args ) == 2 : #the arguments must have a list with 2 arrays having 4 numbers
                if ( ( len( args[0] ) == 4 ) and ( len( args[1] ) == 4 ) ) : 
                   dic["caw"] = args[0][0]
                   dic["cax"] = args[0][1]
                   dic["cay"] = args[0][2]
                   dic["caz"] = args[0][3]
                   dic["cbw"] = args[1][0]
                   dic["cbx"] = args[1][1]
                   dic["cby"] = args[1][2]
                   dic["cbz"] = args[1][3]
                else : raise MovableObjectError, "Object could not oriented with cubic interpolation because args must have two arrays each having four values got '"+str( args[0] )+"' and '"+str( args[1] )+"'"
            else : raise MovableObjectError, "Object could not oriented with cubic interpolation because args must have two arrays each having four values got '"+str( args )+"'"
        self.__setProperty ( self.__MovableObjectProperty_Orientation, dic )
    def rotate ( self, w, x, y, z, time=None, args=None ):
        """Rotates an object. Behaves exactly as setOrientation"""
        self.setOrientation( w, x, y, z, time, args )
    def getOrientation ( self, world=False) :
        """Returs the movable objects orientation in the form of list having four values being [w, x, y, z]. 
        If world is true then we will get the orientation after all applied orientation to parents. 
        If world is false then we get the orientation in relation to parent."""
        result = []
        scale = self.__getProperty ( self.__MovableObjectProperty_Orientation, {"world":str(world).lower()} )
        for i in scale.split( " " ) : result.append ( float( i ) )
        return result

    def getYaw ( self ): 
        """Returns movable objects 'degree' around y axis"""
        dic = {}
        dic["Name"] = self.id
        return float( self.__getProperty ( self.__MovableObjectProperty_Yaw, dic ) )
    
    def setYaw ( self, degree ) :
        """Rotates object 'degree' around y axis"""
        dic = {}
        dic["Name"] = self.id
        dic["Degree"] = degree
        self.__setProperty ( self.__MovableObjectProperty_Yaw, dic )
    def getRoll ( self ): 
        """Returns movable objects 'degree' around z axis"""
        dic = {}
        dic["Name"] = self.id        
        return float( self.__getProperty ( self.__MovableObjectProperty_Roll, dic ) )
    def setRoll ( self, degree ) :
        """Rotates object 'degree' around z axis"""
        dic = {}
        dic["Name"] = self.id
        dic["Degree"] = degree
        self.__setProperty ( self.__MovableObjectProperty_Roll, dic )
        
    def getPitch ( self ): 
        """Returns movable objects 'degree' around x axis"""
        dic = {}
        dic["Name"] = self.id        
        return float( self.__getProperty ( self.__MovableObjectProperty_Pitch, dic ) )
        
    def setPitch ( self, degree ) :
        """Rotates object 'degree' around x axis"""
        dic = {}
        dic["Name"] = self.id
        dic["Degree"] = degree
        self.__setProperty ( self.__MovableObjectProperty_Pitch, dic )
        
    def attach (self, movableobject, position=[]) :
        """ 
        Attaches the 'movableobjectname' to this movable object. The object must exists and be loaded into ogre.
        When attaching an object it will detach the objects parent scene node from its parent. Then it will
        be attached to this objects parent scene node (inheriting its translations + applying its current transitions).
        Cameras are not supported.
        """
        dic = {}
        dic["Name"] = self.id
        dic["SubObjectName"] = movableobject.id
        self.__setProperty ( self.__MovableObjectProperty_Attach, dic )        
        if self.__children == [] : self.__children = []
        self.__children.append(movableobject.id)
        
    def getChildren (self) : return self.__children
        
#NOT TESTED        
#    def lookAt (self, target, enable=True) :
#        """Looks at another movable object 'target' with name 'name'"""
#        dic = {}
#        dic["Name"] = self.id
#        dic["TargetName"] = str(target.id)
#        dic["Enabled"] = str(enable).lower()
#        self.__setProperty (self.__MovableObjectProperty_LookAt, dic)

    def getScale ( self ) :
        """Returns movable objects scale in the form of a list having three float values [x, y, z]"""
        result = []
        scale = self.__getProperty ( self.__MovableObjectProperty_Scale, {} )
        for i in scale.split( " " ) : result.append ( float( i ) )
        return result
    
    def setScale ( self, x, y, z, relative=False ) :
        """Scales relative to current scale"""
        if not validposlist( [x, y, z] ) : raise MovableObjectError, "Scaling does not accept any negative numbers (" + str( [x, y, z] )+")"
        dic = {}
        dic["Name"] = self.id
        dic["ScaleX"] = x
        dic["ScaleY"] = y
        dic["ScaleZ"] = z
        dic["Relative"] = str( relative ).lower()
        self.__setProperty ( self.__MovableObjectProperty_Scale, dic )
        
    def scale ( self, x, y, z ) :
        """Scales relative to current scale"""
        self.setScale ( x, y, z, True )

    def setVisible ( self, b=True, cascade=True ) :
        """if b is True then we set this object visable and all its children"""
        dic = {}
        dic["Name"] = self.id
        dic["Visible"] = str( b ).lower()
        dic["Cascade"] = str ( cascade ).lower()
        self.__setProperty ( self.__MovableObjectProperty_Visible, dic )

    def showBoundingbox ( self, b=True ) :
        """if b is true then we display this object bounding box"""
        dic = {}
        dic["Name"] = self.id
        dic["DisplayBoundingBox"] = str( b ).lower()
        self.__setProperty ( self.__MovableObjectProperty_BoundingBox, dic )

#NOT TESTED
#    def attach (self, sceneNode) :
#        gameGraphics.attachSceneNode (self.id, sceneNode.id)
#
#    def addEventListener (self, listener) :    
#        """ where listener is a event handler type that supports movable objects """
#        gameGraphics.addEventListener(self.id, listener.id )

    def __setProperty ( self, attrib, dic ):
        """Private call that works as a trap into the graphics system handling movable objects, setting values"""
        dic["Name"] = self.id
        if gameGraphics.MovableObject_setProperty ( attrib, str( dic ) ) == 0 :
            raise MovableObjectError, "Could not set property " + str( attrib ) + " using dictionary '" + str( dic ) + "', faild setting state with supplied values"

    def __getProperty ( self, attrib, dic ):
        """Private call that works as a trap into the graphics system handling movable objects, retreaving values"""
        dic["Name"] = self.id
        result = gameGraphics.MovableObject_getProperty ( attrib, str( dic ) )
        if result == None :
            raise MovableObjectError, "Could not get property " + str( attrib ) + " using dictionary '" + str( dic ) + "', faild setting state with supplied values"
        else : return result

    def __str__ ( self ) :
        """Formated string describing name and position of the object"""
        return self.id + ":" + " Position:" + str( self.getPosition() )

##############################################
#    Entity
#        Defines an instance of a discrete, movable object based on a Mesh.
##############################################
class Entity ( MovableObject ) :
    """Handles a mesh, use SceneManager addEntity instead of calling this directly"""
    __EntityProperty_Material = 1
    __EntityProperty_AnimationPlay = 2
    file = ""
    __last_played_animation = None

    def __init__( self, name, file = None ) :
        """This class should not be instanced directly use SceneManager.addEntity(...)"""
        self.id = name
        self.file = file
        if file == None : return None
        if not self._exists () : raise EntityError, "Could not create entity named '" + name + "' because it already exists, use SceneManager to add and remove objects"

    def getAllAnimationNames ( self ) :
        """Returns a list of all the animations supported by this entity."""
        res = gameGraphics.Entity_getAnimationNames ( self.id )
        return res.split( ";" )

    def playAnimation ( self, name, loop=False, enabled=True ) :
        """Plays entity animation and returns None on success. Else raises an EntityError."""
        exists = False
        existing = self.getAllAnimationNames()
        for a in existing : 
            if a == name : exists = True
        if self.__last_played_animation <> None : self.stopAnimation (self.__last_played_animation) #for sound stuff
        if not exists : raise EntityError, "Could not play animation named '"+name+"' because it does not exist in entity animation set "+str( existing )
        else : 
            self.__last_played_animation = name
            dic = {"Name":self.id, "AnimationName":name, "AnimationLoop":str( loop ).lower(), "AnimationEnabled":str( enabled ).lower()}
            anilength = self.__setProperty ( self.__EntityProperty_AnimationPlay, dic )
            if anilength == 0 : raise EntityError, "Could not play existing animation named '"+name+"' because playing it failed"
            else : return anilength

    def stopAnimation ( self, name) :
        """Plays entity animation and returns None on success. Else raises an EntityError."""
        exists = False
        existing = self.getAllAnimationNames()
        for a in existing : 
            if a == name : exists = True
        if not exists : raise EntityError, "Could not stop animation named '"+name+"' because it does not exist in entity animation set "+str( existing )
        else : 
            dic = {"Name":self.id, "AnimationName":name, "AnimationLoop":str( False ).lower(), "AnimationEnabled":str( False ).lower()}
            anilength = self.__setProperty ( self.__EntityProperty_AnimationPlay, dic )
            if anilength == 0 : raise EntityError, "Could not stop existing animation named '"+name+"' because stopping it failed"
            else : return anilength
        
    def setMaterial ( self, MaterialName ) :
        """Sets this entites material. Replaces the material currently in use."""
        self.__setProperty( self.__EntityProperty_Material, {"MaterialName":MaterialName} )
            
    def __setProperty ( self, attrib, dic ) :
        dic["Name"] = self.id
        ret = gameGraphics.Entity_setProperty ( attrib, str( dic ) )
        if ret == 0 : raise EntityError, "Could not set property " + str( attrib ) + " using dictionary '" + str( dic ) + "', faild setting state with supplied values"
        if ret == None : raise EntityError, "Fatal error could not set property " + str( attrib ) + " using dictionary '" + str( dic ) + "', the error is due to invalid types"
        return ret        
        
    def __str__( self ):
        """Formated string describing name, file, animation names and position of the Entity"""
        return "Entity name '"+self.id+"' based on file '"+str( self.file )+"' having animations "+str( self.getAllAnimationNames() )+" at location "+str( self.getPosition() )

##############################################
#    Light
#        Representation of a dynamic light source in the scene
##############################################
class Light ( MovableObject ) :
    __LightProperty_Create = 0
    __LightProperty_DiffuseColor = 1
    __LightProperty_Direction = 2
    __LightProperty_Falloff = 3
    __type = None
    
    def __init__( self, name ) :
        """This class should not be instanced directly use SceneManager.addLight(...)"""
        self.id = name
        if not self._exists () : raise LightError, "Could not create light named '" + name + "' because it already exists, use SceneManager to add and remove objects"

    def setDiffuseColor ( self, r, g, b ) :
        if not validcolor ( r, g, b ) : raise LightError, "Could not set light values to "+str( [r, g, b] )+" because they where not valid color values. Valid color values are unit based values"
        """Sets the colour of the diffuse light given off by this source."""
        self.__setProperty( self.__LightProperty_DiffuseColor, {"Red":r, "Green":g, "Blue":b} )
        
    def setDirection ( self, x, y, z ) :
        """Sets the light to directional light mode and direction in which a light points."""
        self.__setProperty( self.__LightProperty_Direction, {"DirX":x, "DirY":y, "DirZ":z} )

    def setFalloffRange ( self, innerAngle, outerAngle, falloff ) :
        """
        Sets the light to spotlight mode.
        Sets the falloff between the inner and outer cones of the spotlight.
        Where:
            innerAngle is the angle in degrees covered by the bright inner cone The inner cone applicable only to Direct3D, it'll always treat as zero in OpenGL.
            outerAngle is the angle in degrees covered by the outer cone
            falloff is the rate of falloff between the inner and outer cones. 1.0 means a linear falloff, less means slower falloff, higher means faster falloff.
        """
        if not validrange ( innerAngle, 0, 360 ) : raise LightError, "Could not set light values of innerAngle to "+str( innerAngle )+" as it must be possitive degree within 360"
        if not validrange ( outerAngle, innerAngle, 360 ) : raise LightError, "Could not set light values of outerAngle to "+str( outerAngle )+" as it must be larger than innerAngle and within 360"
        if not validrange ( falloff, 0, 1 ) : raise LightError, "Could not set light values of falloff to "+str( falloff )+" as it must be between 0 and 1"
        self.__setProperty( self.__LightProperty_Falloff, {"InnerAngle":innerAngle, "OuterAngle":outerAngle, "Falloff":falloff} )

    def __setProperty ( self, attrib, dic ) :
        dic["Name"] = self.id
        ret = gameGraphics.Light_setProperty ( attrib, str( dic ) )
        if ret == 0 : raise LightError, "Could not set property " + str( attrib ) + " using dictionary '" + str( dic ) + "', faild setting state with supplied values"
        if ret == None : raise LightError, "Fatal error could not set property " + str( attrib ) + " using dictionary '" + str( dic ) + "', the error is due to invalid types"
        return ret        
        
    def __str__( self ):
        """Formated string describing name, file, animation names and position of the Light"""
        return "Light name '"+self.id+"' based of type '"+str( self.__type )+"' at location "+str( self.getPosition() )

##############################################
#    Organic
#        A object that interacts with its enviroment
##############################################
class Organic :

    __type = None
    __dic = {}
    __typeParam = {}

    def __init__( self, name, type, dic ) :
        """This class should not be instanced directly use SceneManager.addOrganic(...)"""
        self.__typeParam["Grass"] = ["GrassWidth", "GrassHeight", "Material", "Radius", "EdgeDistribution", "PositionX", "PositionZ"]
        self.__typeParam["WoblyMesh"] = ["px", "py", "pz", "sz", "sy", "sz", "ow", "ox", "oy", "oz", "Displacement", "Density", "TimeDensity"]
        self.id = name
        self.__type = type
        if ( self.__typeParam.has_key( type ) ) :
            for i in self.__typeParam[type] : 
                if not dic.has_key( i ) : raise OrganicError, "Could not set the parameters of organic type '"+type+"' expected '"+str( self.__typeParam[type] )+"' got '"+str( dic )+"'"
        else :
            dic["Name"] = name
            dic["Type"] = type                
            self.__dic = dic
    def getPropertyKeys ( self, type ): 
        """Returns all the keys needed to be set by the specified type to be a successfull call to the system."""
        if self.__typeParam.has_key( type ) :
            return self.__typeParam[type].keys()
        return None
    def __setProperty ( self, attrib, dic ) :
        dic["Name"] = self.id
        ret = gameGraphics.Organic_setProperty ( attrib, str( dic ) )
        if ret == 0 : raise OrganicError, "Could not set property " + str( attrib ) + " using dictionary '" + str( dic ) + "', faild setting state with supplied values"
        if ret == None : raise OrganicError, "Fatal error could not set property " + str( attrib ) + " using dictionary '" + str( dic ) + "', the error is due to invalid types"
        return ret
    def getType ( self ) : 
        """Returns the Organic type. The supported types are Grass and WoblyMesh"""
        return self.__type;
    def __str__( self ):
        return "Organics name '"+self.id+"' using material '"+str( self.__properties )+"'"

##############################################
#    Particles
#        Class defining particle system based special effects
##############################################
class Particles ( MovableObject ) :
    """Class defining particle system based special effects"""
    __ParticleProperty_Create = 0
    
    def __init__( self, name, effectName ) :
        """This class should not be instanced directly use SceneManager.addParticles(...)"""
        self.id = name
        if not self._exists () : raise ParticleError, "Could not create particle object named '" + name + "' because it already exists, use SceneManager to add and remove objects"

    def __setProperty ( self, attrib, dic ) :
        dic["Name"] = self.id
        if gameGraphics.Particle_setProperty ( attrib, str( dic ) ) == 0 :
            raise ParticleError, "Could not set property " + str( attrib ) + " using dictionary '" + str( dic ) + "', faild setting state with supplied values"

##############################################
#    Animation
#        An animation sequence.
##############################################
class Animation :
    """
    Animation class is a placeholder for animating movable objects. Many movable objects can be 
    animated by the animation class by adding them using Animation.addTrack (...) and supplying 
    the class of the object extended by MovableObject class.
    """
    name = None
    __attrib = None
    __AnimationProperty_Create = 0
    __AnimationProperty_Destroy = 1
    __AnimationProperty_InterpolationMode = 4
    __AnimationProperty_Enabled = 10
    __AnimationProperty_TimePosition = 11
    __AnimationProperty_Loop = 12
    __AnimationProperty_Weight = 13    
    __animationTracks = None

    def __init__( self, name, length ) :
        """
        This class should not be instanced directly use SceneManager.addAnimation(...)
        """
        self.__attrib = {}
        self.__attrib["AnimationName"] = name
        self.__attrib["AnimationLength"] = length
        self.id = name 
        self.__animationTracks = []
        
    def setInterpolationMode ( self, bUsingInterpolation=True ) :
        """ 
        Creating animation class we use interpolation by default. This is changed if bUsingInterpolation is False,
        then we will use Spline to interpolate the animation frames.
        """
        if bUsingInterpolation : self.__attrib["InterpolationMode"] = "Linear"
        else : self.__attrib["InterpolationMode"] = "Spline"
        self.__setProperty ( self.__AnimationProperty_InterpolationMode, self.__attrib )
    
    def play ( self, bLoop=True ) : 
        """Plays the animation. If bLoop equals true then we loop the animation"""
        self.__attrib["Enabled"] = "true"
        if ( bLoop ) : self.__attrib["Loop"] = "true"
        else : self.__attrib["Loop"] = "false"
        self.__setProperty ( self.__AnimationProperty_Enabled, self.__attrib )
        self.__setProperty ( self.__AnimationProperty_Loop, self.__attrib )
    def stop ( self ) : 
        """Stops the animation."""
        self.__attrib["Enabled"] = "false"
        self.__attrib["TimePosition"] = 0
        self.__setProperty ( self.__AnimationProperty_Enabled, self.__attrib )
        self.__setProperty ( self.__AnimationProperty_TimePosition, self.__attrib )
        
    def pause ( self ) : 
        """Pauses the animation."""
        self.__attrib["Enabled"] = "false"
        self.__setProperty ( self.__AnimationProperty_Enabled, self.__attrib )
    
    def setWeight ( self, w ):
        """Sets the weight (influence) of this animation."""
        self.__attrib["Weight"] = w
        self.__setProperty ( self.__AnimationProperty_Weight, self.__attrib )
    
    def addTrack ( self, movableObject ) :
        """
        Creates a new AnimationTrack automatically associated with a object extending MovableObject 'movableObject'.
        This method creates a standard AnimationTrack, but also associates it with a target MovableObject which will receive all keyframe effects. 
        Returns the instance of the create AnimationTrack class that knows its index. Where the index is a positive numeric value from 0 to n.
        """
        self.__animationTracks.append( AnimationTrack ( len( self.__animationTracks ), self.__attrib, movableObject ) )
        return self.__animationTracks[len( self.__animationTracks )-1]
        
    def __setProperty ( self, property, attribute ) :
        if gameGraphics.Graphics_Animation_setProperty ( property, str( attribute ) ) == None :
            raise AnimationError, "Could not set animation property with supplied values '" + str( attribute ) + "' for property '" + str( property ) + "'"

##############################################
#    AnimationTrack
#        'track' in an animation sequence, ie a sequence of keyframes 
#        which affect a certain type of animable object. 
##############################################
class AnimationTrack :
    """
    AnimationTrack is the class that holds a many KeyFrames. The animation track handles the animation
    for the supplied movable object.
    Use Animation class to create animation tracks. Animation can have zero to many AnimationTracks.
    To remove animation track call SceneManager.delAnimation(...).
    """
    
    __keyFrames = None
    index = None
    __attrib = None
    __AnimationProperty_CreateTrack = 2
        
    def __init__( self, index, attrib, movableObject ) :
        """This class should not be instanced directly use SceneManager.addAnimation(...) and then Animation.addTrack(...)"""
        self.__attrib = attrib
        self.__attrib["MovableObjectName"] = movableObject.id
        self.__attrib["TrackHandle"] = index
        self.__setProperty ( self.__AnimationProperty_CreateTrack, attrib ) #pass all know values from Animation class to create the AnimationTrack
        self.__keyFrames = []
        self.index = index
    
    def addKeyFrame ( self, timePos ) :
        """
        Creates a key frame for animation of the movable object. Where tiemPos must be a positive integer value representing the given time index.
        It is better to create KeyFrames in time order. Creating them out of order can result in expensive reordering processing. 
        Note that a KeyFrame at time index 0.0 is always created for you, so you don't need to create this one.
        Note that the timePos must also be lower than the total time of animation (animation lenght).
        KeyFrame is used to define the actual scale, translation and rotation in an animation.
        Returns instance of KeyFrame on success but raises AnimationError on failure.
        """
        if timePos < 0 : 
            raise AnimationError, "Could not add key frame to animation '" + self.__attrib["AnimationName"] + "' on track '"+str( self.index )+"' because timePost must be larger than zero value"
        if float(timePos) > float(self.__attrib["AnimationLength"]) : 
            raise AnimationError, "Could not add key frame to animation '" + self.__attrib["AnimationName"] + "' on track '"+str( self.index )+"' because timePost must not exceed animation length '"+str( self.__attrib["AnimationLength"] )+"' we got '"+str(timePos)+"'"
        self.__keyFrames.append ( KeyFrame ( len( self.__keyFrames ), self.__attrib, timePos ) )
        return self.__keyFrames[len( self.__keyFrames )-1]

    def __setProperty ( self, property, attribute ) :
        if gameGraphics.Graphics_Animation_setProperty ( property, str( attribute ) ) == None :
            raise AnimationError, "Could not set animation track property with supplied values '" + str( attribute ) + "' for property '" + str( property ) + "'"


##############################################
#    KeyFrame
#        A key frame in an animation sequence defined by an AnimationTrack
##############################################    
class KeyFrame :
    """
    KeyFrame is the class that positions an animated movable object by setting the subsiquent transition, rotation or scale at
    specified time position.
    Use AnimationTrack class to create key frames. AnimationTrack can have zero to many key frames.
    To remove animation track call SceneManager.delAnimation(...).
    """
    id = 0
    __AnimationProperty_CreateTrackKeyFrame = 5
    __AnimationProperty_DestroyTrackKeyFrame = 6
    __AnimationProperty_TrackKeyFrameTranslate = 7
    __AnimationProperty_TrackKeyFrameScale = 8
    __AnimationProperty_TrackKeyFrameRotation = 9
    __attrib = None
    
    def __init__( self, id, attrib, timePos ) :
        """
        This class should not be instanced directly use SceneManager.addAnimation(...) and then Animation.addTrack(...) 
        and lastly AnimationTrack.addKeyFrame(...)
        """
        self.__attrib = {}
        self.__attrib = attrib
        self.__attrib["KeyFrameTimePosition"] = timePos
        self.__attrib["KeyFrameIndex"] = id
        self.__setProperty ( self.__AnimationProperty_CreateTrackKeyFrame, self.__attrib )

    def setTranslate ( self, x, y, z ) :
        """
        Sets the translation associated with this keyframe. Returns None on success but raises AnimationError on failure
        """
        self.__attrib["KeyFrameTranslate"] = str( x ) + " " + str( y ) + " " + str( z ) 
        self.__setProperty ( self.__AnimationProperty_TrackKeyFrameTranslate, self.__attrib )

    def setScale ( self, x, y, z ) :
        """
        Sets the scaling factor applied by this keyframe to the animable object at it's time index.  Returns None on success 
        but raises AnimationError on failure
        """
        self.__attrib["KeyFrameScale"] = str( x ) + " " + str( y ) + " " + str( z ) 
        self.__setProperty ( self.__AnimationProperty_TrackKeyFrameScale, self.__attrib )

    def setRotation ( self, q, x, y, z ) :
        """
        Sets the rotation applied by this keyframe.  Returns None on success but raises AnimationError on failure
        """
        self.__attrib["KeyFrameRotation"] = str( q ) + " " + str( x ) + " " + str( y ) + " " + str( z ) 
        self.__setProperty ( self.__AnimationProperty_TrackKeyFrameRotation, self.__attrib )
        
    def __setProperty ( self, property, attribute ) :
        if gameGraphics.Graphics_Animation_setProperty ( property, str( attribute ) ) == None :
            raise AnimationError, "Could not set key frame property with supplied values '" + str( attribute ) + "' for property '" + str( property ) + "'"


##############################################
#    Camera
#        Manages the properties of our view of field
##############################################
class Camera :
    """Request for refactoring. Currenlty we handle the camera creation and cleanup. Ogre should do it instead."""
    
    __CameraViewProperty_Position               = 1
    __CameraViewProperty_Orientation            = 2 
    __CameraViewProperty_LookAtEntity           = 4
    __CameraViewProperty_LookAtDirection        = 5
    __CameraViewProperty_FOV                    = 6
    __CameraViewProperty_ClipDistance           = 7
    __CameraViewProperty_TargetSceneNode        = 8
    __CameraViewProperty_AttachTo               = 9
    __CameraViewProperty_ControlKeys            = 10
    __CameraViewProperty_ControlUnit            = 11
    __CameraViewProperty_FollowAttributes       = 12
    __CameraViewProperty_FollowEnabled          = 13
    __CameraViewProperty_Select                 = 14

    def __init__ ( self) :
        """Is meant to give access to the one default camera"""
        pass
        
    def setFOV ( self, fov=0.8 ) :
        """Sets field of view, 0.8 is a good value"""
        if fov <= 0 : raise CameraError, "Camera got bad FOV value of " + str( fov ) + " fov must be greater than 0 and lower than 1"
        self.__setProperty( self.__CameraViewProperty_FOV, {"FOV":fov} )

    def setPosition ( self, x, y, z ):
        """Positions the camera"""
        dic = {"PositionX":x,"PositionY":y,"PositionZ":z}
        self.__setProperty ( self.__CameraViewProperty_Position, dic )

    def setOrientation ( self, w, x, y, z ):
        """Rotates the camera"""
        dic = {"OrientationW":w,"OrientationX":x,"OrientationY":y,"OrientationZ":z}
        self.__setProperty ( self.__CameraViewProperty_Orientation, dic )

    def lookAtDirection ( self, x, y, z ):
        """Turns the camera to the specified direction"""
        dic = {"DirectionX":x,"DirectionY":y,"DirectionZ":z}
        self.__setProperty ( self.__CameraViewProperty_LookAtDirection, dic )
            
    def lookAtEntity ( self, name="" ) :
        """
        Where name is null or a loaded movable object. Critical: pass null value (disabling) before unloading the
        entity. So we are not looking at the deleted entity (will cause crash).
        """
        dic = {"EntityName":name}
        self.__setProperty ( self.__CameraViewProperty_LookAtEntity, dic )
    
    def setTargetEntity (self, name=""):
        """
        Where name is null or a loaded movable object. Critical: pass null value (disabling) before unloading the
        entity. So we are not looking at the deleted entity (will cause crash).
        """
        dic = {"EntityName":name}
        self.__setProperty ( self.__CameraViewProperty_TargetSceneNode, dic )
        
    def followEntity ( self, name="" ) :
        """
        Where name is null or a loaded movable object. Critical: pass null value (disabling) before unloading the
        entity. So we are not following at the deleted entity (will cause crash).        
        """
        self.lookAtEntity( "" )
        self.attach("")
        dic = {"EntityName":name}
        self.__setProperty ( self.__CameraViewProperty_FollowEnabled, dic )
            
    def setFollowProperties ( self, followFactor, followRotationFactor, height, distance, lookAhead=100 ) :
        """Applies the follow scene node factors. How the scene node is followed."""
        dic = {}
        dic["FollowFactor"] = followFactor
        dic["RotationFactor"] = followRotationFactor
        dic["FollowHeight"] = height
        dic["FollowDistance"] = distance
        dic["FollowLookAhead"] = lookAhead
        self.__setProperty ( self.__CameraViewProperty_FollowAttributes, dic )


    def setFlyFreeControlKeys ( self, MoveLeft, MoveRight, MoveForward, MoveBackward, MoveUp, MoveDown, RotateLeft, RotateRight ) :
        """Sets what keys are used when we control the camera in free fly mode."""
        self.controlKeys = {}
        self.controlKeys["ControlKeyTranslateLeft"] = MoveLeft
        self.controlKeys["ControlKeyTranslateRight"]=MoveRight
        self.controlKeys["ControlKeyTranslateFrwd"]  = MoveForward
        self.controlKeys["ControlKeyTranslateBack"] = MoveBackward
        self.controlKeys["ControlKeyTranslateUp"] = MoveUp
        self.controlKeys["ControlKeyTranslateDown"] = MoveDown
        self.controlKeys["ControlKeyRotateLeft"] = RotateLeft
        self.controlKeys["ControlKeyRotateRight"] = RotateRight
        self.__setProperty ( self.__CameraViewProperty_ControlKeys, self.controlKeys )

    def setFlyFreeControl ( self, moveScale=None, rotationScale=None) :
        """
        Will enable fly free camera control for camera. If moveScale and rotationScale is provided
        these values will be used in transition per unit.
        """
        dic = {"x":0} #we add something to avoid empty dictionary
        if moveScale <> None : 
            if moveScale < 0 : raise CameraError, "Move scale can not be an negative value. Got " + str(moveScale)
            dic["FlyFreeTranslate"] = moveScale
        if rotationScale <> None : 
            if rotationScale < 0 : raise CameraError, "Rotation scale can not be an negative value. Got " + str(rotationScale)
            dic["FlyFreeRotate"] = rotationScale
        self.__setProperty ( self.__CameraViewProperty_ControlUnit, dic )
    
    def attach (self, entityName=None):
        """ 
        This will attach the camera to an entity having name entityName, the entity must exist. If no
        entity is defined then the camera is detached and "free".
        When attached you can still control the camera but in the entitys local space. You will be able to 
        set lookAt and lookAtEntity where the camera will follow the attached entity and look at the targets.
        Calling followEntity will detach the camera from the object (restoring it).
        Deleting an entity havning a camera attached will delete the default camera :(.
        """
        if entityName == None :  entityName = ""
        dic = {"AttachToEntity":entityName}
        self.__setProperty ( self.__CameraViewProperty_AttachTo, dic )

    def disable (self) : 
        """ 
        Will remove any behavoirs on camera.
        """
        self.lookAtEntity()

    def useCamera (self, cameraName = None):
        """
        This will change the camera from the current to the specified 'cameraName'. The camera named 'cameraName'
        must exist in the system (you can create through loading external scenes, like through call loadScene(...)).
        Critical: pass null value (disabling) before unloading the scene that created the camera.
        Passing None (or '') will select the default camera.
        If the camera does not exist we raise an SceneError.
        """
        if cameraName == None : cameraName = "";
        self.__setProperty ( self.__CameraViewProperty_Select, {"CameraName":cameraName} )

    def __getProperty ( self, property, attributes ) : 
        result = gameGraphics.Camera_getProperty ( property, str(attributes) )
        if result == None : raise CameraError, "Got error when getting camera property ("+str(property)+" having values " + str(attributes)
        
    def __setProperty ( self, property, attributes ) :
        error = gameGraphics.Camera_setProperty ( property, str(attributes) )
        if error <> None : raise CameraError, "Got error when setting camera property ("+str(property)+" having values " + str(attributes) + ". Reason : " + error

##############################################
#    SceneManager
#        Class setting up and managing scene states
##############################################
__entities = {}     #collection of all our entities in the scene manager module
__lights = {}       #collection of all our lights in the scene manager module
__organics = {}     #collection of all our organics in the scene manager module
__particles = {}    #collection of all our particles in the scene manager module
__resources = {}    #loaded resources locations 
__animations = {}
__cameras = {"_defaultCamera":Camera()}      #collection of all cameras
__blitters = {}
__paths = {}        #all the added paths

def addBlitterTexture (textureName, fileName, resourceGroup, width, height):
    """Creates an blitter texture and adds it to our blitter texture collection"""
    __blitters[textureName] = TextureBlitter(textureName, fileName, resourceGroup, width, height)
    return __blitters[textureName]

def getBlitterTextureNames () : 
    """Return all the names of our created blitter textures"""
    return __blitters.keys()

def delBlitterTexture (name) : 
    """Will remove and destroy the blitter from our collection"""
    if __blitters.has_key(name) :
        gameGraphics.TextureBlitter_destroy (__blitters[name].id)
        del __blitters[name]

def getCamera () :
    """Returns the default camera created by the system."""
    return __cameras["_defaultCamera"]
    
#def addPath (name, spline, materialName, deltaY=0.3, width=20, orientation=[1,0,0,0]) :
#    """
#    Adds a path with name 'name' having spline points and assigned material 'materialName' with orientation 'orientation'.
#    The path will be as close to the ground as possible.
#    Where spline is an list of two valued list ([...,[a,b],...]) representing the spline data.
#    materialName must be an loaded material accessable to ogre.
#    The deltaY is the distance from the ground.
#    The orientation is used when we use markers for start locations so that we can use the markers rotation.
#    Raises SceneError on failure.
#    Returns None on success.
#    """
#    __PathProperty_Create = 0
#    __PathProperty_AddPoint = 1
#    __PathProperty_Flush = 2
#
#    if len(spline) == 0 : raise SceneError, "Could not create track because there is no spline data"
#    for t in spline : 
#        if len(t) <> 2 :  raise SceneError, "Could not create track because an item in the list had more or less then two values, got " + str (spline)
#    
#    dic = {"Name":name, "MaterialName":materialName}
#    dic["DeltaY"] = deltaY
#    dic["PathWidth"] = width
#    dic["OrientationW"] = orientation[0]
#    dic["OrientationX"] = orientation[1]
#    dic["OrientationY"] = orientation[2]
#    dic["OrientationZ"] = orientation[3]    
#    
#    error = gameGraphics.Path_setProperty ( __PathProperty_Create, str(dic) )
#    if  error <> None : raise SceneError, "SceneManager could not add '" + name + "' got error: " + str(error)
#    for t in spline : 
#        error = gameGraphics.Path_setProperty ( __PathProperty_AddPoint, str({"Name":name, "PointX":t[0], "PointZ":t[1]}) )
#        if  error <> None : raise SceneError, "SceneManager could not create spline'" + name + "' got error: " + str(error)
#    error = gameGraphics.Path_setProperty ( __PathProperty_Flush, str({"Name":name}) )
#    if  error <> None : raise SceneError, "SceneManager could not generate '" + name + "' got error: " + str(error)
#    __paths[name] = {"spline":spline, "material":materialName}    
#    return __paths[name]
#
#def delPath (name) :
#    """
#    Destroys a path with name 'name'
#    Raises SceneError on failure.
#    Returns None on success.
#    """
#    __PathProperty_Destroy = 3
#    if not __paths.has_key(name) : raise SceneError, "Could not delete path '"+str(name)+"'because it does not exist"
#    error = gameGraphics.Path_setProperty ( __PathProperty_Destroy, str({"Name":name}) )
#    if  error <> None : raise SceneError, "SceneManager could not add '" + name + "' got error: " + str(error)
#    del __paths[name]
    
def addEntity ( name, file ) :
    """Adds a entity with name 'name' using mesh resource file 'file'.
    The entity can only be removed by calling SceneManager.delEntity(<name>).
    Returns a instance on success or raises an SceneError if it already exists or anything else went wrong.
    """
    if __entities.has_key ( name ) : raise SceneError, "SceneManager could not add '" + name + "' because it already exists"
    dic = {}
    dic["Name"] = name
    dic["MeshFile"] = file
    if gameGraphics.Entity_setProperty ( 0, str( dic ) ) == 0 : raise SceneError, "SceneManager could not add '" + name + "' because it exists in graphics core"
    __entities[name] = Entity ( name, file )
    return __entities[name]
def delEntity ( name ) :
    """Removes a entity with name 'name' from the system. Must have been created using call addEntity (...).
    Returns None on success or raises an SceneError if it does not exist"""
    if name <> None :
        if __entities.has_key ( name ) == 0: raise SceneError, "SceneManager could not delete '" + name + "' because it does not exists"
        dic = {}
        dic["Name"] = name
        r = gameGraphics.MovableObject_setProperty ( 0, str( dic ) )
        if r == None : raise SceneError, "SceneManager critical error setting property"
        if r == 0 : raise SceneError, "SceneManager could not delete '" + name + "' because it does not exists in graphics core"
        del __entities[name]
def getEntity ( name ) :
    """Returns the instance of entity having name 'name'. Returns a instance or raises an SceneError if it does not exist"""
    if __entities.has_key ( name ) == 0: raise SceneError, "SceneManager could not retreave '" + name + "' because it does not exists"
    return __entities[name]
def hasEntity ( name ) :
    """Returns a true if entity with name 'name' exists."""
    return __entities.has_key(name)    
def getEntityNames () :
    """Returns a list of all entities currently loaded in the system. Returns a list."""
    return __entities.keys()

def addLight ( name, type = 0 ) :
    """Adds a light with name 'name'. The light is returned on success else SceneError is raised. 
    The light can only be removed by calling SceneManager.delLight(<name>).
    Returns a instance on success or raises an SceneError if it already exists or anything else went wrong.
    Different types of light include the following
    * Point light sources give off light equally in all directions, so require only position not direction.
    * Directional lights simulate parallel light beams from a distant source, hence have direction but no position.
    * Spotlights simulate a cone of light from a source so require position and direction, plus extra values for falloff.
    By default adding a light is a point light. Setting a direction by calling setDirection converts it to a directional light. 
    If setting a fall off using the method setFalloffRange (...) then converts it to a spot light.
    """
    if __lights.has_key ( name ) : raise SceneError, "SceneManager could not add '" + name + "' because it already exists"
    if gameGraphics.Light_setProperty ( 0, str( {"Name":name} ) ) == 0 : raise SceneError, "SceneManager could not add '" + name + "' because it exists in graphics core"
    __lights[name] = Light ( name )
    return __lights[name]
def delLight ( name ) :
    """Removes a light with name 'name' from the system. Must have been created using call addLight (...)
    Returns None on success or raises an SceneError if it does not exist
    """
    if name <> None :
        if __lights.has_key ( name ) == 0: raise SceneError, "SceneManager could not delete '" + name + "' because it does not exists"
        dic = {}
        dic["Name"] = name
        r = gameGraphics.MovableObject_setProperty ( 0, str( dic ) )
        if r == None : raise SceneError, "SceneManager critical error setting property"
        if r == 0 : raise SceneError, "SceneManager could not delete '" + name + "' because it does not exists in graphics core"
        del __lights[name]
def getLight ( name ) :
    """Returns the instance of entity having name 'name'"""    
    if __lights.has_key ( name ) == 0: raise SceneError, "SceneManager could not retreave '" + name + "' because it does not exists"
    return __lights[name]
def hasLight ( name ) :
    """Returns a true if light with name 'name' exists"""
    return __lights.has_key(name)    
def getLightNames () :
    """Returns a list of all lights currently loaded in the system"""
    return __lights.keys()

def addOrganics ( name, type, dic ) :
    """Adds a organic with name 'name' having type 'type'. Where type is either 'Grass' or 'WoblyMesh'. 
    The organic using the following attributes passed as a dictionary:
    Grass     => "GrassWidth", "GrassHeight", "Material", "Radius", "EdgeDistribution", "PositionX", "PositionZ"
    WoblyMesh => "MeshName","px","py","pz","sz","sy","sz","ow","ox","oy","oz","Displacement","Density","TimeDensity"
    The entity is returned on success else SceneError is raised. The entity can only be removed by calling SceneManager.delOrganics(<name>).
    Returns a instance on success or raises an SceneError if it already exists or anything else went wrong
    """
    addCmdType = {"Grass":0, "WoblyMesh":3}
    __typeParam = {}
    __typeParam["Grass"] = ["GrassWidth", "GrassHeight", "Material", "Radius", "EdgeDistribution", "PositionX", "PositionZ"]
    __typeParam["WoblyMesh"] = ["MeshName", "px", "py", "pz", "sz", "sy", "sz", "ow", "ox", "oy", "oz", "Displacement", "Density", "TimeDensity"]
    if ( __typeParam.has_key( type ) ) :
        for i in __typeParam[type] : 
            if not dic.has_key( i ) : raise OrganicError, "Could not set the parameters of organic type '"+type+"' expected '"+str( __typeParam[type] )+"' got '"+str( dic )+"'"
        if __organics.has_key ( name ) : raise SceneError, "SceneManager could not add '" + name + "' because it already exists"
        dic["Name"] = name
        dic["Type"] = type
        if gameGraphics.Organic_setProperty ( addCmdType[type], str( dic ) ) == 0 : raise SceneError, "SceneManager could not add '" + name + "' because it exists in graphics core"
        __organics[name] = Organic ( name, type, dic )
        return __organics[name]
    else : raise OrganicError, "Could not add organic type '"+type+"' because it is not supported"
def delOrganics ( name ) :
    """Removes a organics with name 'name' from the system. Must have been created using call addOrganics (...). 
    Returns None on success or raises an SceneError if it does not exist or system failed to remove it"""
    delCmdType = {"Grass":2, "WoblyMesh":5}
    if name <> None :
        if __organics.has_key ( name ) == 0: raise SceneError, "SceneManager could not delete '" + name + "' because it does not exists"
        dic = {}
        dic["Name"] = name
        r = gameGraphics.Organic_setProperty ( delCmdType[__organics[name].getType()], str( dic ) )
        if r == None : raise SceneError, "SceneManager critical error setting property"
        if r == 0 : raise SceneError, "SceneManager could not delete '" + name + "' because it does not exists in graphics core"
        del __organics[name]    
def getOrganics ( name ) :
    """Returns the instance of organic having name 'name'. Returns a instance or raises an SceneError if it does not exist"""    
    if __organics.has_key ( name ) == 0: raise SceneError, "SceneManager could not retreave '" + name + "' because it does not exists"
    return __organics[name]
def hasOrganics( name ) :
    """Returns a true if organics with name 'name' exists"""
    return __organics.has_key(name)    
def getOrganicNames () :
    """Returns a list of all organics currently loaded in the system. Returns a list"""
    return __organics.keys()

def addParticles ( name, file ) :
    """Adds a particle system named 'name' base on file 'file' where the file is of type .particle and is available in Ogre resource library. 
    Returns a instance on success or raises an SceneError if it already exists or anything else went wrong."""
    if __particles.has_key ( name ) : raise SceneError, "SceneManager could not add '" + name + "' because it already exists"
    dic = {}
    dic["Name"] = name
    dic["ParticleFile"] = file
    if gameGraphics.Particle_setProperty ( 0, str( dic ) ) == 0 : raise SceneError, "SceneManager could not add '" + name + "' because it exists in graphics core"
    __particles[name] = Particles ( name, file )
    return __particles[name]
def delParticles ( name ) :
    """Removes a particles with name 'name' from the system. Must have been created using call addParticles (...)
    Returns None on success or raises an SceneError if it does not exist
    """
    if name <> None :
        if __particles.has_key ( name ) == 0: raise SceneError, "SceneManager could not delete '" + name + "' because it does not exists"
        dic = {}
        dic["Name"] = name
        r = gameGraphics.MovableObject_setProperty ( 0, str( dic ) )
        if r == None : raise SceneError, "SceneManager critical error setting property"
        if r == 0 : raise SceneError, "SceneManager could not delete '" + name + "' because it does not exists in graphics core"
        del __particles[name]    
def getParticles ( name ) :
    """Returns the instance of particles having name 'name'. Returns a instance or raises an SceneError if it does not exist"""    
    if __particles.has_key ( name ) == 0: raise SceneError, "SceneManager could not retreave '" + name + "' because it does not exists"
    return __particles[name]
def hasParticles ( name ) :
    """Returns true if the particles with name 'name' exists"""
    return __particles.has_key(name) 
def getParticleNames () :
    """Returns a list of all particles currently loaded in the system. Returns a list"""
    return __particles.keys()

def addAnimation ( name, length=None ) :
    """
    Creates an animation which can be used to animate movable objects. Passing length=None tells the system to use an already defined animation.
    Remarks:
    An animation is a collection of 'tracks' which over time change the position / orientation of Node objects. In this case, the animation will 
    likely have tracks to modify the position / orientation of SceneNode objects, e.g. to make objects move along a path. 
    You don't need to use an Animation object to move objects around - you can do it yourself using the methods of the Node in your FrameListener class. 
    However, when you need relatively complex scripted animation, this is the class to use since it will interpolate between keyframes for you and 
    generally make the whole process easier to manage. 
    A single animation can affect multiple Node objects (each AnimationTrack affects a single Node). In addition, through animation blending a single Node 
    can be affected by multiple animations, athough this is more useful when performing skeletal animation (see Skeleton::createAnimation). 
    Note that whilst it uses the same classes, the animations created here are kept separate from the skeletal animations of meshes (each Skeleton owns those animations). 
    Parameters:
    name is the name of the animation, should be unique within it's parent (e.g. Skeleton)
    length is the length of the animation in seconds. Passing None tels the system that we are going to use an already defined animation (we just want the handle)
    Returns instance of the created Animation class on success but raises SceneError on failure.
    """
    if name <> None :
        if __animations.has_key ( name ) : raise SceneError, "SceneManager could not add '" + name + "' because it already exists"
        if length <> None :
            if length <= 0 : raise SceneError, "SceneManager can not set animation length to '" + str( length ) + "'. Accepts only possitve values larger than one."
        if gameGraphics.Graphics_Animation_setProperty ( 0, str( {"AnimationName":name, "AnimationLength":length} ) ) == None :
            raise SceneError, "SceneManager could not add '" + name + "' because it exists in graphics core"
        else :
            __animations[name] = None #reserve the space
            __animations[name] = Animation ( name, length )
            return __animations[name]
def delAnimation ( name ) :
    """
    Destroys an animation with name 'name'. The entire tracks and keyframes will be removed from the system. Any movable object will not be effected though.
    Returns None on success but raises SceneError on failure.
    """
    if not __animations.has_key( name ) : raise SceneError, "Could not remove animation object '"+name+"' because it does not exist"
    else : 
        del __animations[name]
        if gameGraphics.Graphics_Animation_setProperty ( 1, str( {"AnimationName":name} ) ) == None : raise SceneError, "Could not remove animation object '"+name+"' from core system due to failure"

def hasAnimation ( name ) :
    """Returns true if the animation with name 'name' exists"""
    return __animations.has_key(name) 
    
def getAnimationNames () :
    """Returns a list of all lights currently loaded in the system"""
    return __animations.keys()
    
def getAnimation( name ) :
    """Returns an animation having name 'name'. Raises SceneError if the animation does not exist."""
    if not __animations.has_key( name ) : raise SceneError, "The animation with name '"+name+"' does not exists"
    return __animations[name]

#scene private trap constants
__GraphicsProperty_AmbientLight = 0
__GraphicsProperty_ShadowDistance = 1
__GraphicsProperty_SkyBox = 2
__GraphicsProperty_Scene = 3
__GraphicsProperty_Fog = 4
__GraphicsProperty_Compositor = 5
__GraphicsProperty_Caelum = 6
__GraphicsProperty_PagingTrees = 7
__GraphicsProperty_PagingGrass = 8
__GraphicsProperty_QueryContent = 9

__SceneManagerDictionary = {}

def setAmbientLight ( r, g, b ) :
    """Sets the scenes ambient light where r,g,b are values between 1 and 0. Returns None"""
    dic = {}
    if validcolor( r, g, b ) == False : raise SceneError, "Can only set scene ambient light using real numbers from 0 to 1"
    dic["Red"] = r
    dic["Green"] = g
    dic["Blue"] = b
    _setProperty( __GraphicsProperty_AmbientLight, dic )
    
def setSkyBox ( strMaterialName = None, bEnable = True, scale = 50 ) :
    """
    Sets the scenes skybox. strMaterialName is a loaded resource material defined as a cube texture. Returns None
    If strMaterialName is set to an valid material and bEnable is true then we will texture the sky.
    If strMaterialName is set to None and bEnable is true then we will do a procedural sky.
    If strMaterialName is set to None and bEnable is false then we will make the sky black.
    """
    if not validrange ( scale, 1, 99999999 ) : raise SceneError, "Could not create skybox with defined scale="+str( scale )+". Valid scale is values 1 to 99999999. Recomend using 50"
    dic = {}
    dic["Enabled"] = str( bEnable ).lower()
    dic["MaterialName"] = strMaterialName
    dic["SceneScale"] = scale
    if strMaterialName == None and bEnable : _setProperty( __GraphicsProperty_Caelum, dic )
    else : 
        if strMaterialName == None : dic["MaterialName"] = ""
        _setProperty( __GraphicsProperty_SkyBox, dic )

def addTrees () :
    _setProperty( __GraphicsProperty_PagingTrees, {} )

def addGrass () :
    _setProperty( __GraphicsProperty_PagingGrass, {} )

def loadScene ( fileName=None ) :
    """
    Loads a scene of supported formats. The files must be accessable to the Ogre resource loader.
    Supported types are .osm (oFusion files created through the oFusion max plugin for Ogre3d). 
    Passing None or '' will unload the prevously loaded scene. Only one scene can be loaded at a time.
    Raises SceneError on failure.
    """
    dic = {}
    if fileName == None : fileName = ""
    dic["SceneFile"] = fileName
    _setProperty( __GraphicsProperty_Scene, dic )

def setShadowDistance ( far, fadestart=0, fadeend=0 ) :
    """Sets the shadow distance where """
    dic = {}
    if not validposreal( fadestart ) : raise SceneError, "Can only set shadow fadestart using proportianal values (real values from 0 to 1)"
    if not validposreal( fadeend ) : raise SceneError, "Can only set shadow fadestart using proportianal values (real values from 0 to 1)"
    if ( far < 1 ) : raise SceneError, "Can only set shadow far distance using positive value"
    dic["ShadowFar"] = far                     #Gets the maximum distance away from the camera that shadows will be visible.
    dic["ShadowFadeStart"] = fadestart         #Sets the proportional distance at which texture shadows begin to fade out.
    dic["ShadowFadeEnd"] = fadeend             #Sets the proportional distance at which texture shadows finish to fading out.
    _setProperty( __GraphicsProperty_ShadowDistance, dic )

def setFog ( r, g, b, density=0.001, fogmode=3 ) :
    """Sets fog in the scene where fogmodes are (None=0,exp=1,exp2=2,linear=3)"""
    dic = {}
    if not validcolor( r, g, b ) : raise SceneError, "Can only set scene fog color using real numbers from 0 to 1"
    if not validposreal( density ) : raise SceneError, "Can only set fog density using possitive real values from 0 to 1"    
    if not validrange( fogmode, 0, 3 ) : raise SceneError, "Can only set fog mode using possitive values from 0 to 3"    
    dic["Red"] = r
    dic["Green"] = g
    dic["Blue"] = b
    dic["Density"] = density
    dic["FogMethod"] = fogmode
    _setProperty( __GraphicsProperty_Fog, dic )
    
def setCompostionEffect ( effect ) :
    """Sets the composition effect (last processing of the image) to effect where effect names can be one of defined effects in file .compositor (Bloom, B&W ....)"""
    b = True
    if effect == None :
        if not __SceneManagerDictionary.has_key( "CompositorName" ) : return ;  #exit because we do not have to disable it at startup
        b = False
    else : __SceneManagerDictionary["CompositorName"] = effect
    __SceneManagerDictionary["CompositorEnable"] = str( b ).lower()
    _setProperty ( __GraphicsProperty_Compositor, __SceneManagerDictionary )

def getContentQuery (filter=None):
    """
    Returns a dictionary on the scene manager content. Current supported filters are:
    None : returns everything
    The returned dictionary has keys 'Light', 'Camera', 'Animation', 'Entity'
    """
    return eval(_getProperty (__GraphicsProperty_QueryContent, {"Filter":''}))

def _setProperty ( attrib, dic ) :
    if gameGraphics.Graphics_setProperty ( attrib, str( dic ) ) == 0 :
        raise SceneError, "Could not set property " + str( attrib ) + " using dictionary '" + str( dic ) + "', faild setting state with supplied values"
    
def _getProperty (attrib, dic ) :
    res = gameGraphics.Graphics_getProperty ( attrib, str( dic ) ) 
    if res <> None : return res
    else : raise SceneError, "Could not get property " + str( attrib ) + " using dictionary '" + str( dic ) + "', faild setting state with supplied values"
    
def _cleanUp () :
    """performs house keeping duties"""
    list = getEntityNames()
    for e in list : 
        delEntity( e ) #cleanup all the entities
    list = getLightNames()
    for l in list : 
        delLight( l ) #cleanup all the lights
    list = getOrganicNames()
    for b in list : 
        delOrganics( b ) #cleanup all the billboards
    list = getParticleNames()
    for p in list : 
        delParticles( p ) #cleanup all the particles
    list = getAnimationNames()
    for p in list : 
        delAnimation( p ) #cleanup all the particles    
    list = getBlitterTextureNames()
    for p in list : 
        delBlitterTextureName( p ) #cleanup all the particles    
    list = None
    setCompostionEffect( None )

#def remResourceLocation (self, resourceName) :
#    if self.__resources.has_key(resourceName) :
#        gameGraphics.Resource_unload (resourceName) 
#        del self.__resources[resourceName]
#        print "Resource group removed :", resourceName
#    
#def addResourceLocation (self, resourceName, resourcePath) :
#    #check if resource exists, if not load it
#    if self.__resources.has_key(resourceName) == false:
#        self.__resources[resourceName] = resourcePath
#        gameGraphics.Resource_load (resourcePath, resourceName, "FileSystem")
#        print "Resource group added :", resourceName, resourcePath
#    else :
#        print "Resource group exists:", resourceName, resourcePath
